# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import enum
import re
import xml.etree.ElementTree as ElementTree

import pytest

import psrpcore.types._enum as ps_enum
from psrpcore.types import (
    PSInt,
    PSInt64,
    PSNoteProperty,
    PSObject,
    PSString,
    PSType,
    PSUInt,
)

from ..conftest import COMPLEX_ENCODED_STRING, COMPLEX_STRING, deserialize, serialize


@pytest.mark.parametrize("rehydrate", [True, False])
def test_ps_enum(rehydrate):
    type_name = "MyEnumRehydrated" if rehydrate else "MyEnum"

    @PSType(type_names=[f"System.{type_name}"], rehydrate=rehydrate)
    class EnumTest(ps_enum.PSEnumBase):

        none = 0
        Value1 = 1
        Value2 = 2
        Value3 = 3

    assert str(EnumTest.none) == "EnumTest.none"
    assert repr(EnumTest.none) == "<EnumTest.none: 0>"
    assert str(EnumTest.Value1) == "EnumTest.Value1"
    assert repr(EnumTest.Value1) == "<EnumTest.Value1: 1>"
    assert str(EnumTest.Value2) == "EnumTest.Value2"
    assert str(EnumTest.Value3) == "EnumTest.Value3"

    val = EnumTest.Value1
    assert isinstance(val, PSObject)
    assert isinstance(val, enum.Enum)
    assert isinstance(val, ps_enum.PSEnumBase)
    assert not isinstance(val, PSInt)
    assert isinstance(val.value, PSInt)
    assert isinstance(val, int)

    element = serialize(val)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == f'<Obj RefId="0">'
        f"<I32>1</I32>"
        f'<TN RefId="0">'
        f"<T>System.{type_name}</T>"
        f"<T>System.Enum</T>"
        f"<T>System.ValueType</T>"
        f"<T>System.Object</T>"
        f"</TN>"
        f"<ToString>Value1</ToString>"
        f"</Obj>"
    )

    actual = deserialize(element)
    base_types = [f"System.{type_name}", "System.Enum", "System.ValueType", "System.Object"]

    if rehydrate:
        assert actual == val
        assert str(actual) == "EnumTest.Value1"
        assert isinstance(actual, int)
        assert isinstance(actual, PSObject)
        assert not isinstance(actual, PSInt)
        assert isinstance(actual, ps_enum.PSEnumBase)
        assert isinstance(actual, EnumTest)
        assert isinstance(actual, enum.Enum)
        assert actual.PSTypeNames == base_types

    else:
        # Without hydration we just get the primitive value back
        base_types = [f"Deserialized.{t}" for t in base_types]
        assert actual == val.value
        assert str(actual) == "Value1"
        assert isinstance(actual, int)
        assert isinstance(actual, PSObject)
        assert isinstance(actual, PSInt)
        assert not isinstance(actual, ps_enum.PSEnumBase)
        assert not isinstance(actual, EnumTest)
        assert not isinstance(actual, enum.Enum)
        assert actual.PSTypeNames == base_types


@pytest.mark.parametrize("rehydrate", [True, False])
def test_ps_enum_unsigned_type(rehydrate):
    type_name = "EnumUIntRehydrated" if rehydrate else "EnumUInt"

    @PSType(type_names=[f"System.{type_name}"], rehydrate=rehydrate)
    class EnumTest(ps_enum.PSEnumBase, base_type=PSUInt):

        none = 0
        Value1 = 1
        Value2 = 2
        Value3 = 3

    assert str(EnumTest.none) == "EnumTest.none"
    assert repr(EnumTest.none) == "<EnumTest.none: 0>"
    assert str(EnumTest.Value1) == "EnumTest.Value1"
    assert repr(EnumTest.Value1) == "<EnumTest.Value1: 1>"
    assert str(EnumTest.Value2) == "EnumTest.Value2"
    assert str(EnumTest.Value3) == "EnumTest.Value3"

    val = EnumTest.Value1
    assert isinstance(val, PSObject)
    assert isinstance(val, enum.Enum)
    assert isinstance(val, ps_enum.PSEnumBase)
    assert not isinstance(val, PSUInt)
    assert isinstance(val.value, PSUInt)
    assert isinstance(val, int)

    element = serialize(val)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == f'<Obj RefId="0">'
        f"<U32>1</U32>"
        f'<TN RefId="0">'
        f"<T>System.{type_name}</T>"
        f"<T>System.Enum</T>"
        f"<T>System.ValueType</T>"
        f"<T>System.Object</T>"
        f"</TN>"
        f"<ToString>Value1</ToString>"
        f"</Obj>"
    )

    actual = deserialize(element)
    base_types = [f"System.{type_name}", "System.Enum", "System.ValueType", "System.Object"]

    if rehydrate:
        assert actual == val
        assert str(actual) == "EnumTest.Value1"
        assert isinstance(actual, int)
        assert isinstance(actual, PSObject)
        assert not isinstance(actual, PSUInt)
        assert isinstance(actual, ps_enum.PSEnumBase)
        assert isinstance(actual, EnumTest)
        assert isinstance(actual, enum.Enum)
        assert actual.PSTypeNames == base_types

    else:
        # Without hydration we just get the primitive value back
        base_types = [f"Deserialized.{t}" for t in base_types]
        assert actual == val.value
        assert str(actual) == "Value1"
        assert isinstance(actual, int)
        assert isinstance(actual, PSObject)
        assert isinstance(actual, PSUInt)
        assert not isinstance(actual, ps_enum.PSEnumBase)
        assert not isinstance(actual, EnumTest)
        assert not isinstance(actual, enum.Enum)
        assert actual.PSTypeNames == base_types


@pytest.mark.parametrize("rehydrate", [True, False])
def test_ps_enum_extended_properties(rehydrate):
    type_name = "EnumExtendedRehydrated" if rehydrate else "EnumExtended"

    @PSType(type_names=[f"System.{type_name}"], rehydrate=rehydrate)
    class EnumTest(ps_enum.PSEnumBase, base_type=PSInt64):

        none = 0
        Value1 = 1
        Value2 = 2
        Value3 = 3

    assert str(EnumTest.none) == "EnumTest.none"
    assert repr(EnumTest.none) == "<EnumTest.none: 0>"
    assert str(EnumTest.Value1) == "EnumTest.Value1"
    assert repr(EnumTest.Value1) == "<EnumTest.Value1: 1>"
    assert str(EnumTest.Value2) == "EnumTest.Value2"
    assert str(EnumTest.Value3) == "EnumTest.Value3"

    val = EnumTest.none
    val.PSObject.extended_properties.append(PSNoteProperty(COMPLEX_STRING))
    val[COMPLEX_STRING] = COMPLEX_STRING
    assert isinstance(val, PSObject)
    assert isinstance(val, enum.Enum)
    assert isinstance(val, ps_enum.PSEnumBase)
    assert not isinstance(val, PSInt64)
    assert isinstance(val.value, PSInt64)
    assert isinstance(val.value, int)

    element = serialize(val)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == f'<Obj RefId="0">'
        f"<I64>0</I64>"
        f'<TN RefId="0">'
        f"<T>System.{type_name}</T>"
        f"<T>System.Enum</T>"
        f"<T>System.ValueType</T>"
        f"<T>System.Object</T>"
        f"</TN>"
        f"<MS>"
        f'<S N="{COMPLEX_ENCODED_STRING}">{COMPLEX_ENCODED_STRING}</S>'
        f"</MS>"
        f"<ToString>None</ToString>"
        f"</Obj>"
    )

    actual = deserialize(element)
    base_types = [f"System.{type_name}", "System.Enum", "System.ValueType", "System.Object"]

    assert val[COMPLEX_STRING] == COMPLEX_STRING

    if rehydrate:
        assert actual == val
        assert str(actual) == "EnumTest.none"
        assert isinstance(actual, int)
        assert isinstance(actual, PSObject)
        assert not isinstance(actual, PSInt64)
        assert isinstance(actual, ps_enum.PSEnumBase)
        assert isinstance(actual, EnumTest)
        assert isinstance(actual, enum.Enum)
        assert actual.PSTypeNames == base_types

    else:
        # Without hydration we just get the primitive value back
        base_types = [f"Deserialized.{t}" for t in base_types]
        assert actual == val.value
        assert str(actual) == "None"
        assert isinstance(actual, int)
        assert isinstance(actual, PSObject)
        assert isinstance(actual, PSInt64)
        assert not isinstance(actual, ps_enum.PSEnumBase)
        assert not isinstance(actual, EnumTest)
        assert not isinstance(actual, enum.Enum)
        assert actual.PSTypeNames == base_types


@pytest.mark.parametrize("rehydrate", [True, False])
def test_ps_flags(rehydrate):
    type_name = "FlagHydrated" if rehydrate else "Flag"

    @PSType(type_names=[f"System.{type_name}"], rehydrate=rehydrate)
    class FlagTest(ps_enum.PSFlagBase):

        none = 0
        Flag1 = 1
        Flag2 = 2
        Flag3 = 4

    assert str(FlagTest.none) == "FlagTest.none"
    assert repr(FlagTest.none) == "<FlagTest.none: 0>"
    assert str(FlagTest.Flag1) == "FlagTest.Flag1"
    assert repr(FlagTest.Flag1) == "<FlagTest.Flag1: 1>"
    assert str(FlagTest.Flag2) == "FlagTest.Flag2"
    assert str(FlagTest.Flag3) == "FlagTest.Flag3"
    assert str(FlagTest.Flag1 | FlagTest.Flag3) == "FlagTest.Flag3|Flag1"
    assert repr(FlagTest.Flag1 | FlagTest.Flag3) == "<FlagTest.Flag3|Flag1: 5>"

    val = FlagTest.Flag1 | FlagTest.Flag3
    assert isinstance(val, PSObject)
    assert isinstance(val, enum.Flag)
    assert not isinstance(val, ps_enum.PSEnumBase)
    assert isinstance(val, ps_enum.PSFlagBase)
    assert not isinstance(val, PSInt)
    assert isinstance(val.value, PSInt)
    assert isinstance(val, int)

    element = serialize(val)

    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == f'<Obj RefId="0">'
        f"<I32>5</I32>"
        f'<TN RefId="0">'
        f"<T>System.{type_name}</T>"
        f"<T>System.Enum</T>"
        f"<T>System.ValueType</T>"
        f"<T>System.Object</T>"
        f"</TN>"
        f"<ToString>Flag1, Flag3</ToString>"
        f"</Obj>"
    )

    actual = deserialize(element)
    base_types = [f"System.{type_name}", "System.Enum", "System.ValueType", "System.Object"]

    if rehydrate:
        assert actual == val
        assert str(actual) == "FlagTest.Flag3|Flag1"
        assert isinstance(actual, int)
        assert isinstance(actual, PSObject)
        assert not isinstance(actual, PSInt)
        assert isinstance(actual, ps_enum.PSFlagBase)
        assert isinstance(actual, FlagTest)
        assert isinstance(actual, enum.Flag)
        assert actual.PSTypeNames == base_types

    else:
        # Without hydration we just get the primitive value back
        base_types = [f"Deserialized.{t}" for t in base_types]
        assert actual == val.value
        assert str(actual) == "Flag1, Flag3"
        assert isinstance(actual, int)
        assert isinstance(actual, PSObject)
        assert isinstance(actual, PSInt)
        assert not isinstance(actual, ps_enum.PSFlagBase)
        assert not isinstance(actual, FlagTest)
        assert not isinstance(actual, enum.Flag)
        assert actual.PSTypeNames == base_types

    element = serialize(FlagTest.none)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == f'<Obj RefId="0">'
        f"<I32>0</I32>"
        f'<TN RefId="0">'
        f"<T>System.{type_name}</T>"
        f"<T>System.Enum</T>"
        f"<T>System.ValueType</T>"
        f"<T>System.Object</T>"
        f"</TN>"
        f"<ToString>None</ToString>"
        f"</Obj>"
    )

    element = serialize(FlagTest.none | FlagTest.Flag2)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == f'<Obj RefId="0">'
        f"<I32>2</I32>"
        f'<TN RefId="0">'
        f"<T>System.{type_name}</T>"
        f"<T>System.Enum</T>"
        f"<T>System.ValueType</T>"
        f"<T>System.Object</T>"
        f"</TN>"
        f"<ToString>Flag2</ToString>"
        f"</Obj>"
    )


def test_ps_flags_operators():
    @PSType(type_names=["System.FlagTest"])
    class FlagTest(ps_enum.PSFlagBase):

        none = 0
        Flag1 = 1
        Flag2 = 2
        Flag3 = 4
        Flag4 = 8

    val = FlagTest.none
    assert val == FlagTest.none
    assert val != FlagTest.Flag1
    assert str(val) == "FlagTest.none"
    assert val.name == "none"
    assert val.value == 0

    val |= FlagTest.Flag1 | FlagTest.Flag2
    assert isinstance(val, FlagTest)
    assert str(val) == "FlagTest.Flag2|Flag1"
    assert val.name is None
    assert val.value == 3

    val &= FlagTest.Flag1
    assert isinstance(val, FlagTest)
    assert str(val) == "FlagTest.Flag1"
    assert val.name == "Flag1"
    assert val.value == 1

    val = (FlagTest.Flag1 | FlagTest.Flag2) ^ FlagTest.Flag1
    assert isinstance(val, FlagTest)
    assert str(val) == "FlagTest.Flag2"
    assert val.value == 2

    val = val << 2
    assert val == FlagTest.Flag4
    assert str(val) == "FlagTest.Flag4"
    assert val.name == "Flag4"
    assert val.value == 8

    val = val >> 2
    assert val == FlagTest.Flag2
    assert str(val) == "FlagTest.Flag2"
    assert val.name == "Flag2"
    assert val.value == 2

    val = ~val
    assert isinstance(val, FlagTest)
    assert str(val) == "FlagTest.Flag4|Flag3|Flag1"
    assert val.name is None
    assert val.value == -3


def test_ps_enum_not_inheriting_int_base():
    expected = re.escape("PSEnumType InvalidEnum base_type must be a subclass of PSIntegerBase")
    with pytest.raises(TypeError, match=expected):

        @PSType(type_names=["Test"])
        class InvalidEnum(ps_enum.PSEnumBase, base_type=PSString):
            none = 0


def test_ps_enum_to_ps_ps_baseint():
    @PSType(type_names=["System.EnumToInt"])
    class EnumToInt(ps_enum.PSEnumBase):

        none = 0
        Value1 = 1

    value = PSInt(EnumToInt.Value1)
    assert isinstance(value, PSInt)
    assert value == 1

    value = PSInt64(EnumToInt.Value1)
    assert isinstance(value, PSInt64)
    assert value == 1



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import uuid

from psrpcore import _payload as payload
from psrpcore.types import PSRPMessageType


def test_unpack_message_without_bom():
    actual = payload.unpack_message(bytearray(b"\x01\x00\x00\x00\x02\x00\x01\x00" + (b"\x00") * 32 + b"abc"))
    assert isinstance(actual, payload.Message)
    assert actual.destination == 1
    assert actual.message_type == PSRPMessageType.SessionCapability
    assert actual.rpid == uuid.UUID(int=0)
    assert actual.pid is None
    assert actual.data == bytearray(b"abc")


def test_unpack_message_with_bom():
    actual = payload.unpack_message(
        bytearray(b"\x01\x00\x00\x00\x02\x00\x01\x00" + (b"\x00") * 32 + b"\xEF\xBB\xBFabc")
    )
    assert isinstance(actual, payload.Message)
    assert actual.destination == 1
    assert actual.message_type == PSRPMessageType.SessionCapability
    assert actual.rpid == uuid.UUID(int=0)
    assert actual.pid is None
    assert actual.data == bytearray(b"abc")



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import re
import xml.etree.ElementTree as ElementTree

import pytest

import psrpcore.types._base as ps_base
from psrpcore.types import PSCustomObject, PSInt, PSString, PSUInt

from ..conftest import deserialize, serialize


def test_ps_note_property_set_value():
    prop = ps_base.PSNoteProperty("Property", ps_type=PSInt)
    prop.set_value("1", None)
    actual = prop.get_value(None)
    assert actual == 1
    assert isinstance(actual, PSInt)

    prop.set_value(None, None)
    assert prop.get_value(None) is None

    expected = re.escape("invalid literal for int() with base 10: 'a'")
    with pytest.raises(ValueError, match=expected):
        prop.set_value("a", None)


def test_ps_property_with_value_and_getter():
    class MasterProperty(ps_base.PSPropertyInfo):
        def copy(self):
            return self

    with pytest.raises(ValueError, match="Cannot set property value for 'Name' with a getter"):
        MasterProperty("Name", value="value", getter=lambda s: None)


def test_ps_property_set_getter_with_value():
    prop = ps_base.PSNoteProperty("Name", "value")

    with pytest.raises(ValueError, match="Cannot add getter for 'Name': existing value already set"):
        prop.getter = lambda s: None


def test_ps_property_set_getter_as_none():
    prop = ps_base.PSScriptProperty("Name", lambda s: None)

    with pytest.raises(ValueError, match="Cannot unset property getter for 'Name'"):
        prop.getter = None


def test_ps_property_unset_setter():
    prop = ps_base.PSScriptProperty("Name", lambda s: None, lambda s, v: None)
    prop.setter = None

    with pytest.raises(ValueError, match="Cannot set value for a getter property 'Name' without a setter callable"):
        prop.set_value("test", None)


def test_ps_property_getter_as_none():
    with pytest.raises(TypeError, match="Cannot create script property 'Name' with getter as None"):
        ps_base.PSScriptProperty("Name", None)


def test_ps_property_no_setter_but_mandatory():
    expected = "Cannot create mandatory PSScriptProperty property 'Name' without a setter callable"
    with pytest.raises(TypeError, match=expected):
        ps_base.PSScriptProperty("Name", lambda s: None, mandatory=True)


def test_ps_script_property_invalid_getter():
    def no_params():
        assert False

    def two_params(this, other):
        assert False

    def args_params(this, other, *args):
        assert False

    expected = re.escape(
        "Invalid getter callable for property 'Callable': signature expected 1 parameter but 0 " "parameters were found"
    )
    with pytest.raises(TypeError, match=expected):
        ps_base.PSScriptProperty("Callable", no_params)

    expected = re.escape(
        "Invalid getter callable for property 'Callable': signature expected 1 parameter but 2 "
        "required parameters were found"
    )
    with pytest.raises(TypeError, match=expected):
        ps_base.PSScriptProperty("Callable", two_params)

    expected = re.escape(
        "Invalid getter callable for property 'Callable': signature expected 1 parameter but 2 "
        "required parameters were found"
    )
    with pytest.raises(TypeError, match=expected):
        ps_base.PSScriptProperty("Callable", args_params)

    with pytest.raises(TypeError, match=expected):
        ps_base.PSScriptProperty("Callable", lambda x, y: None)

    expected = re.escape("Invalid getter callable for property 'Callable': expecting callable not str")
    with pytest.raises(TypeError, match=expected):
        ps_base.PSScriptProperty("Callable", "string")


def test_ps_script_property_invalid_setter():
    def no_params():
        assert False

    def one_param(this):
        assert False

    def only_kwargs(this, *, key=None):
        assert False

    def three_params(this, other, third):
        assert False

    expected = re.escape(
        "Invalid setter callable for property 'Callable': signature expected 2 parameters but 0 "
        "parameters were found"
    )
    with pytest.raises(TypeError, match=expected):
        ps_base.PSScriptProperty("Callable", one_param, no_params)

    expected = re.escape(
        "Invalid setter callable for property 'Callable': signature expected 2 parameters but 1 " "parameter were found"
    )
    with pytest.raises(TypeError, match=expected):
        ps_base.PSScriptProperty("Callable", one_param, one_param)

    with pytest.raises(TypeError, match=expected):
        ps_base.PSScriptProperty("Callable", one_param, only_kwargs)

    expected = re.escape(
        "Invalid setter callable for property 'Callable': signature expected 2 parameters but 3 "
        "required parameters were found"
    )
    with pytest.raises(TypeError, match=expected):
        ps_base.PSScriptProperty("Callable", one_param, three_params)

    with pytest.raises(TypeError, match=expected):
        ps_base.PSScriptProperty("Callable", one_param, lambda x, y, z: None)


def test_ps_script_property_setter_without_getter():
    prop = ps_base.PSNoteProperty("Callable")

    expected = re.escape("Cannot set property setter for 'Callable' without an existing getter")
    with pytest.raises(ValueError, match=expected):
        prop.setter = lambda s, v: None


def test_ps_object_with_script_property():
    @ps_base.PSType(
        type_names=["ScriptablePSObject"],
        extended_properties=[
            ps_base.PSScriptProperty(
                "ScriptMandatory", lambda s: s.world, lambda s, v: setattr(s, "world", v), mandatory=True
            ),
            ps_base.PSNoteProperty("NoteProperty", value="note value"),
            ps_base.PSScriptProperty("ScriptProperty", lambda s: s.test, ps_type=PSInt),
            ps_base.PSScriptProperty(
                "ScriptToNote", lambda s: s.NoteProperty, lambda s, v: setattr(s, "NoteProperty", v)
            ),
        ],
    )
    class ScriptablePSObject(ps_base.PSObject):
        @property
        def test(self):
            return "1"

        @property
        def world(self):
            return getattr(self, "_world", None)

        @world.setter
        def world(self, value):
            setattr(self, "_world", value)

    expected = re.escape("missing 1 required arguments: 'ScriptMandatory'")
    with pytest.raises(TypeError, match=expected):
        ScriptablePSObject()

    obj = ScriptablePSObject("hello")
    assert obj.ScriptProperty == 1
    assert isinstance(obj.ScriptProperty, PSInt)

    element = serialize(obj)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        '<TN RefId="0"><T>ScriptablePSObject</T><T>System.Object</T></TN>'
        "<MS>"
        '<S N="ScriptMandatory">hello</S>'
        '<S N="NoteProperty">note value</S>'
        '<I32 N="ScriptProperty">1</I32>'
        '<S N="ScriptToNote">note value</S>'
        "</MS>"
        "</Obj>"
    )

    obj = deserialize(element)
    assert isinstance(obj, ScriptablePSObject)
    assert isinstance(obj.PSObject.extended_properties[0], ps_base.PSNoteProperty)
    assert isinstance(obj.PSObject.extended_properties[1], ps_base.PSNoteProperty)
    assert isinstance(obj.PSObject.extended_properties[2], ps_base.PSNoteProperty)
    # Not in the CLIXML so the original type was preservedy)
    assert isinstance(obj.PSObject.extended_properties[3], ps_base.PSNoteProperty)

    assert obj.world is None
    assert obj.ScriptMandatory == "hello"
    assert obj.NoteProperty == "note value"
    assert obj.ScriptProperty == 1
    assert isinstance(obj.ScriptProperty, PSInt)
    assert obj.ScriptToNote == "note value"

    obj = ScriptablePSObject("hello", ScriptToNote="other note")
    obj.world = None

    element = serialize(obj)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        '<TN RefId="0"><T>ScriptablePSObject</T><T>System.Object</T></TN>'
        "<MS>"
        '<Nil N="ScriptMandatory" />'
        '<S N="NoteProperty">other note</S>'
        '<I32 N="ScriptProperty">1</I32>'
        '<S N="ScriptToNote">other note</S>'
        "</MS>"
        "</Obj>"
    )


def test_ps_object_with_aliases():
    @ps_base.PSType(
        type_names=["AliasPSObject"],
        extended_properties=[
            ps_base.PSNoteProperty("NoteProperty", mandatory=True),
            ps_base.PSAliasProperty("AliasToNote", "NoteProperty"),
            ps_base.PSAliasProperty("AliasWithType", "NoteProperty", ps_type=PSInt),
            ps_base.PSAliasProperty("AliasToAttr", "test"),
        ],
    )
    class AliasPSObject(ps_base.PSObject):
        def __str__(self):
            return self.__class__.__name__

        @property
        def test(self):
            return "abc"

        @property
        def none(self):
            return getattr(self, "_none", None)

        @none.setter
        def none(self, value):
            setattr(self, "_none", value)

    obj = AliasPSObject("1")
    assert obj.AliasToNote == "1"
    assert obj.AliasWithType == 1
    assert isinstance(obj.AliasWithType, PSInt)
    assert obj.AliasToAttr == "abc"

    element = serialize(obj)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        '<TN RefId="0"><T>AliasPSObject</T><T>System.Object</T></TN>'
        "<MS>"
        '<S N="NoteProperty">1</S>'
        '<S N="AliasToNote">1</S>'
        '<I32 N="AliasWithType">1</I32>'
        '<S N="AliasToAttr">abc</S>'
        "</MS>"
        "<ToString>AliasPSObject</ToString>"
        "</Obj>"
    )

    obj = deserialize(element)
    assert isinstance(obj, AliasPSObject)
    assert isinstance(obj.PSObject.extended_properties[0], ps_base.PSNoteProperty)
    assert isinstance(obj.PSObject.extended_properties[1], ps_base.PSNoteProperty)
    assert isinstance(obj.PSObject.extended_properties[2], ps_base.PSNoteProperty)
    assert isinstance(obj.PSObject.extended_properties[3], ps_base.PSNoteProperty)
    assert obj.NoteProperty == "1"
    assert obj.AliasToNote == "1"
    assert obj.AliasWithType == 1
    assert isinstance(obj.AliasWithType, PSInt)
    assert obj.AliasToAttr == "abc"


def test_ps_object_invalid_init_args():
    @ps_base.PSType(
        type_names=["MyPSObject"],
        adapted_properties=[
            ps_base.PSNoteProperty("Property1"),
        ],
        extended_properties=[
            ps_base.PSNoteProperty("Property2"),
            ps_base.PSNoteProperty("Property3", value="default value"),
        ],
    )
    class MyPSObject(ps_base.PSObject):
        pass

    expected = re.escape("takes 4 positional arguments but 5 were given")
    with pytest.raises(TypeError, match=expected):
        MyPSObject(1, 2, 3, 4)

    expected = re.escape("got multiple values for argument 'Property2'")
    with pytest.raises(TypeError, match=expected):
        MyPSObject(1, 2, Property2="other")

    expected = re.escape("got an unexpected keyword argument 'Other'")
    with pytest.raises(TypeError, match=expected):
        MyPSObject(Other="fake")


def test_ps_object_shadowed_property():
    @ps_base.PSType(
        type_names=["MyPSShadowedObject"],
        adapted_properties=[
            ps_base.PSNoteProperty("Property"),
        ],
        extended_properties=[
            ps_base.PSNoteProperty("Property"),
        ],
    )
    class MyPSShadowedObject(ps_base.PSObject):
        def __str__(self):
            return "mock"

    obj = MyPSShadowedObject("positional")
    assert obj.Property == "positional"
    assert obj.PSObject.adapted_properties[0].get_value(None) is None  # ETS is favoured over Adapted props
    assert obj.PSObject.extended_properties[0].get_value(None) == "positional"

    element = serialize(obj)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0"><TN RefId="0"><T>MyPSShadowedObject</T><T>System.Object</T></TN>'
        '<Props><Nil N="Property" /></Props>'
        '<MS><S N="Property">positional</S></MS>'
        "<ToString>mock</ToString></Obj>"
    )

    obj = deserialize(element)
    assert isinstance(obj, ps_base.PSObject)
    assert isinstance(obj, MyPSShadowedObject)
    assert obj.PSObject.adapted_properties[0].get_value(None) is None
    assert obj.PSObject.extended_properties[0].get_value(None) == "positional"
    assert obj.Property == "positional"
    assert obj.PSTypeNames == ["MyPSShadowedObject", "System.Object"]

    obj = MyPSShadowedObject(Property="kwarg")
    assert obj.Property == "kwarg"
    assert obj.PSObject.adapted_properties[0].get_value(None) is None
    assert obj.PSObject.extended_properties[0].get_value(None) == "kwarg"


def test_ps_object_mandatory():
    @ps_base.PSType(
        type_names=["MandatoryPSObject"],
        extended_properties=[
            ps_base.PSNoteProperty("Mandatory", mandatory=True),
            ps_base.PSNoteProperty("MandatoryTyped", mandatory=True, ps_type=PSUInt),
        ],
    )
    class MandatoryPSObject(ps_base.PSObject):
        def __str__(self):
            # Ensure our serialization has a common value.
            return "mock"

    expected = re.escape("missing 2 required arguments: 'Mandatory', 'MandatoryTyped'")
    with pytest.raises(TypeError, match=expected):
        MandatoryPSObject()

    expected = re.escape("missing 1 required arguments: 'MandatoryTyped'")
    with pytest.raises(TypeError, match=expected):
        MandatoryPSObject("Mandatory")

    # With positional
    obj = MandatoryPSObject("Mandatory", 1)
    assert obj.Mandatory == "Mandatory"
    assert obj.MandatoryTyped == 1
    assert isinstance(obj.MandatoryTyped, PSUInt)

    # With kwargs
    obj = MandatoryPSObject(Mandatory="Mandatory", MandatoryTyped=1)
    assert obj.Mandatory == "Mandatory"
    assert obj.MandatoryTyped == 1
    assert isinstance(obj.MandatoryTyped, PSUInt)

    # With mixture
    obj = MandatoryPSObject("Mandatory", MandatoryTyped=1)
    assert obj.Mandatory == "Mandatory"
    assert obj.MandatoryTyped == 1
    assert isinstance(obj.MandatoryTyped, PSUInt)

    element = serialize(obj)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0"><TN RefId="0"><T>MandatoryPSObject</T><T>System.Object</T></TN>'
        '<MS><S N="Mandatory">Mandatory</S>'
        '<U32 N="MandatoryTyped">1</U32></MS>'
        "<ToString>mock</ToString>"
        "</Obj>"
    )

    obj = deserialize(element)
    assert isinstance(obj, ps_base.PSObject)
    assert isinstance(obj, MandatoryPSObject)
    assert obj.Mandatory == "Mandatory"
    assert isinstance(obj.Mandatory, PSString)
    assert obj.MandatoryTyped == 1
    assert isinstance(obj.MandatoryTyped, PSUInt)
    assert obj.PSTypeNames == ["MandatoryPSObject", "System.Object"]


def test_ps_object_with_init():
    @ps_base.PSType(
        type_names=["ObjectWithInit"],
        extended_properties=[
            ps_base.PSNoteProperty("Property"),
            ps_base.PSNoteProperty("Mandatory", mandatory=True),
        ],
    )
    class ObjectWithInit(ps_base.PSObject):
        def __init__(self, value):
            self.value = value

    # If __init__ is defined on the type then automatic validation and setting of props during init is not done.
    obj = ObjectWithInit("value")
    assert not obj.PSObject.extended_properties[0].mandatory
    assert obj.PSObject.extended_properties[1].mandatory
    assert obj.value == "value"
    assert obj.Property is None
    assert obj.Mandatory is None


def test_ps_object_with_multi_inheritance():
    class OtherObj:
        def __init__(self):
            self.value = "other"

    @ps_base.PSType(
        type_names=["MultiObj"],
        extended_properties=[
            ps_base.PSNoteProperty("Mandatory", mandatory=True),
            ps_base.PSNoteProperty("Property"),
        ],
    )
    class MultiObj(ps_base.PSObject, OtherObj):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            OtherObj.__init__(self)  # Needed to MultiObj __init__ is called.

    # If a class inherits PSObject and another class the automatic __init__ isn't called.
    with pytest.raises(TypeError, match="missing 1 required arguments: 'Mandatory'"):
        MultiObj()

    obj = MultiObj("value")
    assert obj.Mandatory == "value"
    assert obj.Property is None
    assert obj.PSObject.extended_properties[0].mandatory
    assert not obj.PSObject.extended_properties[1].mandatory
    assert obj.value == "other"


def test_ps_object_init_with_nested_inheritance():
    @ps_base.PSType(["EmptyParentPSObject"])
    class EmptyParentPSObject(ps_base.PSObject):
        pass

    @ps_base.PSType(
        type_names=["YounglingPSObject"],
        extended_properties=[
            ps_base.PSNoteProperty("Property"),
            ps_base.PSNoteProperty("Mandatory", mandatory=True),
        ],
    )
    class YounglingPSObject(EmptyParentPSObject):
        pass

    assert YounglingPSObject.PSObject.type_names == ["YounglingPSObject", "EmptyParentPSObject", "System.Object"]
    assert EmptyParentPSObject.PSObject.type_names == ["EmptyParentPSObject", "System.Object"]

    obj = YounglingPSObject("prop", "man")
    assert obj.Property == "prop"
    assert obj.Mandatory == "man"
    assert obj.PSObject.type_names == ["YounglingPSObject", "EmptyParentPSObject", "System.Object"]

    element = serialize(obj)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        '<TN RefId="0"><T>YounglingPSObject</T><T>EmptyParentPSObject</T><T>System.Object</T></TN>'
        "<MS>"
        '<S N="Property">prop</S>'
        '<S N="Mandatory">man</S>'
        "</MS>"
        "</Obj>"
    )

    with pytest.raises(TypeError, match="missing 1 required arguments: 'Mandatory'"):
        YounglingPSObject()


def test_ps_object_init_properties_nested():
    @ps_base.PSType(
        type_names=["Nested1"],
        adapted_properties=[
            ps_base.PSNoteProperty("Adapted1Mandatory", mandatory=True),
            ps_base.PSNoteProperty("Adapted1"),
        ],
        extended_properties=[
            ps_base.PSNoteProperty("Extended1", ps_type=PSInt),
        ],
    )
    class Nested1(ps_base.PSObject):
        pass

    @ps_base.PSType(
        type_names=["Nested2"],
        extended_properties=[
            ps_base.PSNoteProperty("Extended2", mandatory=True),
        ],
    )
    class Nested2(Nested1):
        def __str__(self):
            return self.Adapted1Mandatory

    @ps_base.PSType(
        type_names=["Nested3"],
        extended_properties=[
            ps_base.PSNoteProperty("Extended3"),
        ],
    )
    class Nested3(Nested2):
        pass

    assert Nested1.PSObject.type_names == ["Nested1", "System.Object"]
    assert len(Nested1.PSObject.adapted_properties) == 2
    assert Nested1.PSObject.adapted_properties[0].name == "Adapted1Mandatory"
    assert Nested1.PSObject.adapted_properties[1].name == "Adapted1"
    assert len(Nested1.PSObject.extended_properties) == 1
    assert Nested1.PSObject.extended_properties[0].name == "Extended1"

    assert Nested2.PSObject.type_names == ["Nested2", "Nested1", "System.Object"]
    assert len(Nested2.PSObject.adapted_properties) == 2
    assert len(Nested2.PSObject.extended_properties) == 2
    assert Nested2.PSObject.extended_properties[0].name == "Extended2"

    assert Nested3.PSObject.type_names == ["Nested3", "Nested2", "Nested1", "System.Object"]
    assert len(Nested3.PSObject.adapted_properties) == 2
    assert len(Nested3.PSObject.extended_properties) == 3
    assert Nested3.PSObject.extended_properties[0].name == "Extended3"

    expected = re.escape("missing 2 required arguments: 'Adapted1Mandatory', 'Extended2'")
    with pytest.raises(TypeError, match=expected):
        Nested3()

    with pytest.raises(TypeError, match=expected):
        Nested3(Extended3="test")

    obj = Nested3(Adapted1Mandatory="adapted", Extended1="1", Extended2="extended 2")
    assert obj.Adapted1Mandatory == "adapted"
    assert obj.Adapted1 is None
    assert obj.Extended1 == 1
    assert obj.Extended2 == "extended 2"
    assert obj.Extended3 is None
    assert str(obj) == "adapted"

    element = serialize(obj)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        '<TN RefId="0"><T>Nested3</T><T>Nested2</T><T>Nested1</T><T>System.Object</T></TN>'
        "<Props>"
        '<S N="Adapted1Mandatory">adapted</S>'
        '<Nil N="Adapted1" />'
        "</Props>"
        "<MS>"
        '<Nil N="Extended3" />'
        '<S N="Extended2">extended 2</S>'
        '<I32 N="Extended1">1</I32>'
        "</MS>"
        "<ToString>adapted</ToString>"
        "</Obj>"
    )

    obj = deserialize(element)
    assert isinstance(obj, Nested3)
    assert isinstance(obj, Nested2)
    assert isinstance(obj, Nested1)
    assert isinstance(obj, ps_base.PSObject)
    assert str(obj) == "adapted"
    assert obj.Adapted1Mandatory == "adapted"
    assert obj.Adapted1 is None
    assert obj.Extended1 == 1
    assert obj.Extended2 == "extended 2"
    assert obj.Extended3 is None


def test_ps_object_init_with_parent_init_inheritance():
    @ps_base.PSType(["OtherPSObject"])
    class OtherPSObject(ps_base.PSObject):
        def __init__(self, value):
            self.value = value

    @ps_base.PSType(
        type_names=["GrandPSObject"],
        extended_properties=[
            ps_base.PSNoteProperty("Property"),
            ps_base.PSNoteProperty("Mandatory", mandatory=True),
        ],
    )
    class GrandPSObject(OtherPSObject):
        pass

    # Needs to be a direct descendant to get the auto __init__ function
    with pytest.raises(TypeError, match="got an unexpected keyword argument 'Property'"):
        GrandPSObject("other", Property="abc")

    obj = GrandPSObject("other")
    assert not obj.PSObject.extended_properties[0].mandatory
    assert obj.PSObject.extended_properties[1].mandatory
    assert obj.value == "other"
    assert obj.Property is None
    assert obj.Mandatory is None


def test_dynamic_ps_object():
    obj = ps_base.PSObject()
    obj.PSObject.type_names = ["Type1", "Type2"]
    ps_base.add_note_property(obj, "NoteProperty", "note value")
    ps_base.add_alias_property(obj, "AliasProperty", "NoteProperty")

    element = serialize(obj)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        '<TN RefId="0"><T>Type1</T><T>Type2</T></TN>'
        "<MS>"
        '<S N="NoteProperty">note value</S>'
        '<S N="AliasProperty">note value</S>'
        "</MS>"
        "</Obj>"
    )


def test_ps_object_dict():
    obj = ps_base.PSObject()
    obj.PSObject.type_names = ["Type1"]
    ps_base.add_note_property(obj, "NoteProperty", "note value")

    # The raw __dict__ does not contain our note properties, we add that to a copy.
    raw_dict = object.__getattribute__(obj, "__dict__")
    assert list(raw_dict.keys()) == ["PSObject"]

    actual = vars(obj)
    assert id(actual) != id(raw_dict)
    assert list(actual.keys()) == ["PSObject", "NoteProperty"]
    assert actual["NoteProperty"] == "note value"

    # Adding a value to our copy of __dict__ won't affect the obj itself
    actual["other"] = "other value"
    assert "other" not in object.__getattribute__(obj, "__dict__")

    with pytest.raises(AttributeError):
        obj.other


def test_add_member_properties():
    # Set up a base object for up to play with
    obj = PSCustomObject(Existing="value", Integer="1")
    obj.PSObject.adapted_properties.append(ps_base.PSNoteProperty("Adapted", "test"))

    ps_base.add_alias_property(obj, "Alias", "Existing")
    ps_base.add_alias_property(obj, "AliasTyped", "Integer", ps_type=PSInt)
    ps_base.add_note_property(obj, "Note", "note value")
    ps_base.add_script_property(obj, "Script", lambda s: s.Note)
    ps_base.add_script_property(obj, "ScriptSet", lambda s: s.AliasTyped, lambda s, v: setattr(s, "Integer", v))

    assert len(obj.PSObject.adapted_properties) == 1
    assert len(obj.PSObject.extended_properties) == 7
    assert isinstance(obj.PSObject.extended_properties[0], ps_base.PSNoteProperty)
    assert obj.PSObject.extended_properties[0].name == "Existing"
    assert isinstance(obj.PSObject.extended_properties[1], ps_base.PSNoteProperty)
    assert obj.PSObject.extended_properties[1].name == "Integer"
    assert isinstance(obj.PSObject.extended_properties[2], ps_base.PSAliasProperty)
    assert obj.PSObject.extended_properties[2].name == "Alias"
    assert isinstance(obj.PSObject.extended_properties[3], ps_base.PSAliasProperty)
    assert obj.PSObject.extended_properties[3].name == "AliasTyped"
    assert isinstance(obj.PSObject.extended_properties[4], ps_base.PSNoteProperty)
    assert obj.PSObject.extended_properties[4].name == "Note"
    assert isinstance(obj.PSObject.extended_properties[5], ps_base.PSScriptProperty)
    assert obj.PSObject.extended_properties[5].name == "Script"
    assert isinstance(obj.PSObject.extended_properties[6], ps_base.PSScriptProperty)
    assert obj.PSObject.extended_properties[6].name == "ScriptSet"

    assert obj.Alias == "value"
    assert obj.AliasTyped == 1
    assert isinstance(obj.AliasTyped, PSInt)
    assert obj.Note == "note value"
    assert obj.Script == "note value"
    assert obj.ScriptSet == 1
    assert isinstance(obj.ScriptSet, PSInt)

    obj.ScriptSet = "2"
    assert obj.AliasTyped == 2
    assert obj.ScriptSet == 2

    # Make sure the members were only added to that instance
    obj = PSCustomObject()
    assert obj.PSObject.adapted_properties == []
    assert obj.PSObject.adapted_properties == []


def test_add_member_against_existing():
    obj = PSCustomObject(Extended="extended")
    obj.PSObject.adapted_properties.append(ps_base.PSNoteProperty("Adapted", "adapted"))

    assert obj.Adapted == "adapted"
    assert obj.Extended == "extended"

    expected = re.escape("Property 'Adapted' already exists on PSObject, use force=True to overwrite it")
    with pytest.raises(RuntimeError, match=expected):
        ps_base.add_note_property(obj, "Adapted", "new adapted")

    expected = re.escape("Property 'Extended' already exists on PSObject, use force=True to overwrite it")
    with pytest.raises(RuntimeError, match=expected):
        ps_base.add_note_property(obj, "Extended", "new extended")

    ps_base.add_note_property(obj, "Adapted", "new adapted", force=True)
    ps_base.add_note_property(obj, "Extended", "new extended", force=True)

    assert obj.Adapted == "new adapted"
    assert obj.Extended == "new extended"

    assert len(obj.PSObject.adapted_properties) == 1
    assert obj.PSObject.adapted_properties[0].name == "Adapted"
    assert obj.PSObject.adapted_properties[0].get_value(obj) == "adapted"

    assert len(obj.PSObject.extended_properties) == 2
    assert obj.PSObject.extended_properties[0].name == "Extended"
    assert obj.PSObject.extended_properties[0].get_value(obj) == "new extended"
    assert obj.PSObject.extended_properties[1].name == "Adapted"
    assert obj.PSObject.extended_properties[1].get_value(obj) == "new adapted"


def test_add_member_against_object_class():
    @ps_base.PSType(
        type_names=["MutatedClass"],
    )
    class MutatedClass(ps_base.PSObject):
        pass

    first = MutatedClass()
    assert len(first.PSObject.adapted_properties) == 0
    assert len(first.PSObject.extended_properties) == 0

    ps_base.add_alias_property(MutatedClass, "Alias", "Note")
    ps_base.add_note_property(MutatedClass, "Note", "note value")
    ps_base.add_script_property(MutatedClass, "Script", lambda s: s.Alias)

    assert len(MutatedClass.PSObject.adapted_properties) == 0
    assert len(MutatedClass.PSObject.extended_properties) == 3
    assert isinstance(MutatedClass.PSObject.extended_properties[0], ps_base.PSAliasProperty)
    assert MutatedClass.PSObject.extended_properties[0].name == "Alias"
    assert isinstance(MutatedClass.PSObject.extended_properties[1], ps_base.PSNoteProperty)
    assert MutatedClass.PSObject.extended_properties[1].name == "Note"
    assert isinstance(MutatedClass.PSObject.extended_properties[2], ps_base.PSScriptProperty)
    assert MutatedClass.PSObject.extended_properties[2].name == "Script"

    # Verify existing classes aren't mutated
    assert len(first.PSObject.adapted_properties) == 0
    assert len(first.PSObject.extended_properties) == 0

    # Verify new class instances are
    second = MutatedClass()
    assert len(second.PSObject.adapted_properties) == 0
    assert len(second.PSObject.extended_properties) == 3
    assert isinstance(second.PSObject.extended_properties[0], ps_base.PSAliasProperty)
    assert second.PSObject.extended_properties[0].name == "Alias"
    assert isinstance(second.PSObject.extended_properties[1], ps_base.PSNoteProperty)
    assert second.PSObject.extended_properties[1].name == "Note"
    assert isinstance(second.PSObject.extended_properties[2], ps_base.PSScriptProperty)
    assert second.PSObject.extended_properties[2].name == "Script"

    assert second.Alias == "note value"
    assert second.Note == "note value"
    assert second.Script == "note value"


def test_ps_isinstance_from_string():
    ps_string = PSString()

    assert ps_base.ps_isinstance(ps_string, "System.String")
    assert ps_base.ps_isinstance(ps_string, ("System.Int32", "System.String"))
    assert ps_base.ps_isinstance(ps_string, "System.Object")
    assert not ps_base.ps_isinstance(ps_string, "System.Int32")


def test_ps_instance_from_ps_type():
    ps_string = PSString()

    assert ps_base.ps_isinstance(ps_string, PSString)
    assert ps_base.ps_isinstance(ps_string, (PSInt, PSString))
    assert not ps_base.ps_isinstance(ps_string, PSInt)


def test_ps_instance_with_deserialized_object():
    @ps_base.PSType(
        type_names=["Deserialized.My.Other", "Deserialized.Intermediate", "Deserialzied.Object"],
    )
    class MyDeserialized1(ps_base.PSObject):
        pass

    @ps_base.PSType(
        type_names=["Deserialized.Intermediate", "Deserialzied.Object"],
    )
    class MyDeserialized2(ps_base.PSObject):
        pass

    @ps_base.PSType(
        type_names=["My.Other", "Intermediate", "ect"],
    )
    class Serialized(ps_base.PSObject):
        pass

    obj = MyDeserialized1()
    assert ps_base.ps_isinstance(obj, MyDeserialized1)
    assert ps_base.ps_isinstance(obj, MyDeserialized2)
    assert not ps_base.ps_isinstance(obj, Serialized)
    assert ps_base.ps_isinstance(obj, Serialized, ignore_deserialized=True)


def test_fail_to_create_ps_type_without_ps_object():

    with pytest.raises(TypeError, match=r"PSType class [\w_\.<>]+ must be a subclass of PSObject"):

        @ps_base.PSType()
        class MyClass:
            pass


def test_fail_add_member_no_ps_object():
    expected = re.escape("The passing in object does not contain the required PSObject attribute")

    with pytest.raises(ValueError, match=expected):
        ps_base.add_note_property("", "test", "value")



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import base64
import typing
import uuid

from psrpcore._pipeline import GetMetadata, PowerShell
from psrpcore.types import (
    ApartmentState,
    ApplicationPrivateData,
    BufferCell,
    ChoiceDescription,
    ConnectRunspacePool,
    ConsoleColor,
    Coordinates,
    CreatePipeline,
    DebugRecord,
    DebugRecordMsg,
    EncryptedSessionKey,
    EndOfPipelineInput,
    ErrorRecord,
    ErrorRecordMsg,
    FieldDescription,
    GetAvailableRunspaces,
    GetCommandMetadata,
    HostInfo,
    HostMethodIdentifier,
    InformationRecord,
    InformationRecordMsg,
    InitRunspacePool,
    KeyInfo,
    PipelineHostCall,
    PipelineHostResponse,
    PipelineState,
    ProgressRecord,
    ProgressRecordMsg,
    PSBool,
    PSCredentialTypes,
    PSCredentialUIOptions,
    PSInvocationState,
    PSObject,
    PSRPMessageType,
    PSThreadOptions,
    PSVersion,
    PublicKey,
    PublicKeyRequest,
    ReadKeyOptions,
    Rectangle,
    ResetRunspaceState,
    RunspaceAvailability,
    RunspacePoolHostCall,
    RunspacePoolHostResponse,
    RunspacePoolInitData,
    RunspacePoolState,
    RunspacePoolStateMsg,
    SessionCapability,
    SetMaxRunspaces,
    SetMinRunspaces,
    Size,
    UserEvent,
    VerboseRecord,
    VerboseRecordMsg,
    WarningRecord,
    WarningRecordMsg,
)

T1 = typing.TypeVar("T1")
T2 = typing.TypeVar("T2")
_OptionalPipelineType = typing.Optional[uuid.UUID]


def _decode_host_resp_result(
    method: HostMethodIdentifier,
    result: typing.Any,
) -> typing.Optional[typing.Any]:
    if result is None:
        return None

    elif method in [HostMethodIdentifier.GetForegroundColor, HostMethodIdentifier.GetBackgroundColor]:
        return ConsoleColor(result)

    elif method in [HostMethodIdentifier.GetCursorPosition, HostMethodIdentifier.GetWindowPosition]:
        return Coordinates.FromPSObjectForRemoting(result)

    elif method in [
        HostMethodIdentifier.GetBufferSize,
        HostMethodIdentifier.GetWindowSize,
        HostMethodIdentifier.GetMaxWindowSize,
        HostMethodIdentifier.GetMaxPhysicalWindowSize,
    ]:
        return Size.FromPSObjectForRemoting(result)

    elif method == HostMethodIdentifier.ReadKey:
        return KeyInfo.FromPSObjectForRemoting(result)

    elif method == HostMethodIdentifier.GetBufferContents:
        return _unpack_multi_dimensional_array(result, BufferCell.FromPSObjectForRemoting)

    else:
        return result


def _decode_host_call_parameters(
    method: HostMethodIdentifier,
    parameters: typing.List[typing.Any],
) -> typing.List[typing.Any]:
    if method in [HostMethodIdentifier.Write2, HostMethodIdentifier.WriteLine3]:
        return [ConsoleColor(parameters[0]), ConsoleColor(parameters[1]), parameters[2]]

    elif method == HostMethodIdentifier.WriteProgress:
        return [
            parameters[0],
            ProgressRecord(
                Activity=parameters[1].Activity,
                ActivityId=parameters[1].ActivityId,
                CurrentOperation=parameters[1].CurrentOperation,
                ParentActivityId=parameters[1].ParentActivityId,
                PercentComplete=parameters[1].PercentComplete,
                RecordType=parameters[1].Type,
                SecondsRemaining=parameters[1].SecondsRemaining,
                StatusDescription=parameters[1].StatusDescription,
            ),
        ]

    elif method == HostMethodIdentifier.Prompt:
        return [
            parameters[0],
            parameters[1],
            [FieldDescription.FromPSObjectForRemoting(f) for f in parameters[2]],
        ]

    elif method == HostMethodIdentifier.PromptForCredential2:
        return [
            parameters[0],
            parameters[1],
            parameters[2],
            parameters[3],
            PSCredentialTypes(parameters[4]),
            PSCredentialUIOptions(parameters[5]),
        ]

    elif method in [HostMethodIdentifier.PromptForChoice, HostMethodIdentifier.PromptForChoiceMultipleSelection]:
        return [
            parameters[0],
            parameters[1],
            [ChoiceDescription.FromPSObjectForRemoting(c) for c in parameters[2]],
            parameters[3],
        ]

    elif method in [HostMethodIdentifier.SetForegroundColor, HostMethodIdentifier.SetBackgroundColor]:
        return [ConsoleColor(parameters[0])]

    elif method in [HostMethodIdentifier.SetCursorPosition, HostMethodIdentifier.SetWindowPosition]:
        return [Coordinates.FromPSObjectForRemoting(parameters[0])]

    elif method in [HostMethodIdentifier.SetBufferSize, HostMethodIdentifier.SetWindowSize]:
        return [Size.FromPSObjectForRemoting(parameters[0])]

    elif method == HostMethodIdentifier.ReadKey:
        return [ReadKeyOptions(parameters[0])]

    elif method == HostMethodIdentifier.SetBufferContents1:
        return [Rectangle.FromPSObjectForRemoting(parameters[0]), BufferCell.FromPSObjectForRemoting(parameters[1])]

    elif method == HostMethodIdentifier.SetBufferContents2:
        cells = _unpack_multi_dimensional_array(parameters[1], BufferCell.FromPSObjectForRemoting)
        return [Coordinates.FromPSObjectForRemoting(parameters[0]), cells]

    elif method == HostMethodIdentifier.GetBufferContents:
        return [Rectangle.FromPSObjectForRemoting(parameters[0])]

    elif method == HostMethodIdentifier.ScrollBufferContents:
        return [
            Rectangle.FromPSObjectForRemoting(parameters[0]),
            Coordinates.FromPSObjectForRemoting(parameters[1]),
            Rectangle.FromPSObjectForRemoting(parameters[2]),
            BufferCell.FromPSObjectForRemoting(parameters[3]),
        ]

    else:
        return parameters


def _unpack_multi_dimensional_array(
    obj: PSObject,
    unpack_obj: typing.Callable[[PSObject], typing.Any],
) -> typing.List:

    final_list: typing.List[typing.Any] = [unpack_obj(o) for o in obj.mae]
    for count in reversed(obj.mal):
        to_enumerate = final_list
        final_list = []

        entries: typing.Any = []
        for entry in to_enumerate:
            entries.append(entry)
            if len(entries) == count:
                final_list.append(entries)
                entries = []

    return final_list[0]


class PSRPEvent(typing.Generic[T1, T2]):
    """PSRP Event.

    Represents an event based on the various PSRP messages that can be received
    from a peer. It contains information such as the Runspace Pool and Pipeline
    the event is targeted towards as well as the raw PSObject received for
    further parsing.

    Args:
        message_type: The :class:`psrpcore.types.PSRPMessageType` that
            identifies the message type.
        ps_object: The message received from the peer.
        runspace_pool_id: The Runspace Pool the event is for.
        pipeline_id: The Pipeline the event is for, otherwise `None` when the
            event is just for the Runspace Pool.

    Attributes:
        message_type: See args.
        runspace_pool_id: See args.
        pipeline_id: See args.
    """

    def __init__(
        self,
        message_type: PSRPMessageType,
        ps_object: T1,
        runspace_pool_id: uuid.UUID,
        pipeline_id: T2,
    ):
        self.message_type = message_type
        self.runspace_pool_id = runspace_pool_id
        self.pipeline_id = pipeline_id

        self._ps_object = ps_object

    @classmethod
    def create(
        cls,
        message_type: PSRPMessageType,
        ps_object: T1,
        runspace_pool_id: uuid.UUID,
        pipeline_id: T2 = None,
    ) -> "PSRPEvent":
        """Creates an event for the specific message types - used internally."""
        event_cls = _REGISTRY[message_type]
        return event_cls(message_type, ps_object, runspace_pool_id, pipeline_id)

    def __repr__(self) -> str:
        return f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} pipeline_id={self.pipeline_id!r}>"


_REGISTRY: typing.Dict[PSRPMessageType, typing.Type[PSRPEvent]] = {}


def RegisterEvent(cls: typing.Type) -> typing.Type[PSRPEvent]:
    """Registers an event based on the PSRPMessageType."""
    msg_type = cls.__orig_bases__[0].__args__[0]

    # PipelineInput and PipelineOutput can accept anything.
    msg_id = None
    if msg_type != typing.Any:
        msg_id = PSRPMessageType.get_message_id(msg_type)

    elif cls.__module__ in RegisterEvent.__module__:
        msg_id = {
            "PipelineInputEvent": PSRPMessageType.PipelineInput,
            "PipelineOutputEvent": PSRPMessageType.PipelineOutput,
        }[cls.__qualname__]

    if msg_id:
        _REGISTRY[msg_id] = cls

    return cls


@RegisterEvent
class ApplicationPrivateDataEvent(PSRPEvent[ApplicationPrivateData, None]):
    """Application Private Data Event.

    Event Information:

        Direction: Server -> Client

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.ApplicationPrivateData`

        Action Required: Share with higher layer

    The server generates this data automatically when receiving an
    :class:`psrpcore.types.InitRunspacePool` or
    :class:`psrpcore.types.ConnectRunspacePool` message from the client.
    """

    @property
    def data(self) -> typing.Dict[str, typing.Any]:
        """The private data dictionary returned from the server."""
        return self._ps_object.ApplicationPrivateData

    def __repr__(self) -> str:
        return f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r}: {self.data!r}>"


@RegisterEvent
class ConnectRunspacePoolEvent(PSRPEvent[ConnectRunspacePool, None]):
    """Connect Runspace Pool Event.

    Event Information:

        Direction: Client -> Server

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.ConnectRunspacePool`

        Action Required: Send responses back to the client

    The client sends this message when attempting to connect to a disconnected
    Runspace Pool. The server will generate the
    :class:`psrpcore.types.RunspacePoolInitData`,
    :class:`psrpcore.types.ApplicationPrivateData`, and
    :class:`psrpcore.types.RunspacePoolStateMsg` to send back to the client.
    """

    @property
    def max_runspaces(self) -> typing.Optional[int]:
        """The maximum number of runspaces in the Runspace Pool."""
        return getattr(self._ps_object, "MaxRunspaces", None)

    @property
    def min_runspaces(self) -> typing.Optional[int]:
        """The minimum number of runspaces in the Runspace Pool."""
        return getattr(self._ps_object, "MinRunspaces", None)

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} "
            f"min_runspaces={self.min_runspaces} max_runspaces={self.max_runspaces}>"
        )


@RegisterEvent
class CreatePipelineEvent(PSRPEvent[CreatePipeline, uuid.UUID]):
    """Create Pipeline Event.

    Event Information:

        Direction: Client -> Server

        For: Pipeline

        Message Type: :class:`psrpcore.types.CreatePipeline`

        Action Required: Process pipeline

    The client sends this message when invoking a new pipeline. There are no
    responses automatically generated to send back to the client as it is up
    to the server to invoke when there are available resources.
    """

    @property
    def pipeline(self) -> PowerShell:
        """The PowerShell pipeline details."""
        return PowerShell.FromPSObjectForRemoting(self._ps_object)

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} pipeline_id={self.pipeline_id!r} "
            f"pipeline={self.pipeline!r}>"
        )


@RegisterEvent
class DebugRecordEvent(PSRPEvent[DebugRecordMsg, _OptionalPipelineType]):
    """Debug Record Event.

    Event Information:

        Direction: Server -> Client

        For: Runspace Pool or Pipeline

        Message Type: :class:`psrpcore.types.DebugRecordMsg`

        Action Required: Send to higher layer

    The server sends this message when there is a debug record generated by the
    server. The data should be shared with the higher layer as necessary.
    """

    @property
    def record(self) -> DebugRecord:
        """The DebugRecord emitted by the peer."""
        return self._ps_object

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} pipeline_id={self.pipeline_id!r} "
            f"record={self.record.Message!r}>"
        )


@RegisterEvent
class EncryptedSessionKeyEvent(PSRPEvent[EncryptedSessionKey, _OptionalPipelineType]):
    """Encrypted Session Key Event.

    Event Information:

        Direction: Server -> Client

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.EncryptedSessionKey`

        Action Required: None - secure strings can now be serialized

    The server sends this message after processing the
    :class:`psrpcore.types.PublicKey` message from the client. It contains the
    encrypted session key used when serializing
    :class:`psrpcore.types.PSSecureString` objects.
    """

    @property
    def key(self) -> bytes:
        """The session key generated by the peer."""
        return base64.b64decode(self._ps_object.EncryptedSessionKey)


@RegisterEvent
class EndOfPipelineInputEvent(PSRPEvent[EndOfPipelineInput, uuid.UUID]):
    """End Of Pipeline Input Event.

    Event Information:

        Direction: Client -> Server

        For: Pipeline

        Message Type: :class:`psrpcore.types.EndOfPipeline`

        Action Required: Server notifies the pipeline that no more input is to
        be expected.

    The client sends this message once it has finished sending all the input
    for the pipeline.
    """


@RegisterEvent
class ErrorRecordEvent(PSRPEvent[ErrorRecordMsg, _OptionalPipelineType]):
    """Error Record Event.

    Event Information:

        Direction: Server -> Client

        For: Runspace Pool or Pipeline

        Message Type: :class:`psrpcore.types.ErrorRecordMsg`

        Action Required: Send to higher layer

    The server sends this message when there is an error record generated by the
    server. The data should be shared with the higher layer as necessary.
    """

    @property
    def record(self) -> ErrorRecord:
        """The ErrorRecord emitted by the peer."""
        return self._ps_object

    def __repr__(self) -> str:
        err = str(self.record)
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} pipeline_id={self.pipeline_id!r} "
            f"record={err!r}>"
        )


@RegisterEvent
class GetAvailableRunspacesEvent(PSRPEvent[GetAvailableRunspaces, None]):
    """Get Available Runspaces Event.

    Event Information:

        Direction: Client -> Server

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.GetAvailableRunspaces`

        Action Required: Send response back to the client

    The client sends this to request the number of available runspaces in the
    Runspace Pool. The server automatically generates the
    :class:`psrpcore.types.RunspaceAvailability` as a response to the client.
    """

    @property
    def ci(self) -> int:
        """The call id associated with the request."""
        return self._ps_object.ci

    def __repr__(self) -> str:
        return f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} ci={self.ci}>"


@RegisterEvent
class GetCommandMetadataEvent(PSRPEvent[GetCommandMetadata, uuid.UUID]):
    """Get Command Metadata Event.

    Event Information:

        Direction: Client -> Server

        For: Pipeline

        Message Type: :class:`psrpcore.types.GetCommandMetadata`

        Action Required: Process pipeline

    The client sends this message when requesting command metadata in a
    Runspace Pool. There are no responses automatically generated to send back
    to the client as it is up to the server to process when there are available
    resources.
    """

    @property
    def pipeline(self) -> GetMetadata:
        """The GetCommandMetadata pipeline details."""
        return GetMetadata.FromPSObjectForRemoting(self._ps_object)

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} pipeline_id={self.pipeline_id!r} "
            f"pipeline={self.pipeline!r}>"
        )


@RegisterEvent
class InformationRecordEvent(PSRPEvent[InformationRecordMsg, _OptionalPipelineType]):
    """Information Record Event.

    Event Information:

        Direction: Server -> Client

        For: Runspace Pool or Pipeline

        Message Type: :class:`psrpcore.types.InformationRecordMsg`

        Action Required: Send to higher layer

    The server sends this message when there is an information record generated
    by the server. The data should be shared with the higher layer as necessary.
    """

    @property
    def record(self) -> InformationRecord:
        """The InformationRecord emitted by the peer."""
        # The raw PSObject generated in an INFORMATION_RECORD msg looks like a Information record but isn't exactly the
        # same. It serialzies with extended props whereas a proper info record uses adapted properties. By creating
        # the object and just copying the properties over as is it will seem like the object is the same as a normal
        # info record as defined in types.
        info = InformationRecord()
        info.PSObject.adapted_properties = self._ps_object.PSObject.extended_properties

        return info

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} pipeline_id={self.pipeline_id!r} "
            f"record={self.record.MessageData!r}>"
        )


@RegisterEvent
class InitRunspacePoolEvent(PSRPEvent[InitRunspacePool, None]):
    """Init Runspace Pool Event.

    Event Information:

        Direction: Client -> Server

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.InitRunspacePool`

        Action Required: Send responses back to the client

    The client sends this message when attempting to open a new Runspace Pool.
    The server will generate the :class:`psrpcore.types.ApplicationPrivateData`
    and :class:`psrpcore.types.RunspacePoolStateMsg` to send back to the
    client.
    """

    @property
    def max_runspaces(self) -> int:
        """The maximum number of runspaces in the Runspace Pool."""
        return self._ps_object.MaxRunspaces

    @property
    def min_runspaces(self) -> int:
        """The minimum number of runspaces in the Runspace Pool."""
        return self._ps_object.MinRunspaces

    @property
    def ps_thread_options(self) -> PSThreadOptions:
        """The PowerShell thread options specified by the peer."""
        return self._ps_object.PSThreadOptions

    @property
    def apartment_state(self) -> ApartmentState:
        """The apartment state specified by the peer."""
        return self._ps_object.ApartmentState

    @property
    def host_info(self) -> HostInfo:
        """The PSHost info provided by the peer."""
        return HostInfo.FromPSObjectForRemoting(self._ps_object.HostInfo)

    @property
    def application_arguments(self) -> typing.Dict[str, typing.Any]:
        """Higher layer application arguments provided by the peer."""
        return self._ps_object.ApplicationArguments

    def __repr__(self) -> str:
        kw = " ".join(
            [
                f"{k}={getattr(self, k)!r}"
                for k in [
                    "min_runspaces",
                    "max_runspaces",
                    "ps_thread_options",
                    "apartment_state",
                    "host_info",
                    "application_arguments",
                ]
            ]
        )
        return f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} {kw}>"


@RegisterEvent
class PipelineHostCallEvent(PSRPEvent[PipelineHostCall, uuid.UUID]):
    """Pipeline Host Call Event.

    Event Information:

        Direction: Server -> Client

        For: Pipeline

        Message Type: :class:`psrpcore.types.PipelineHostCall`

        Action Required: Send to higher layer for processing

    The server sends this message when requesting the host to invoke a
    :class:`psrpcore.types.HostMethodIdentifier` method. It is up to the client
    to send this information to the higher layer and send back a response if
    required.
    """

    @property
    def ci(self) -> int:
        """The call id associated with the request."""
        return self._ps_object.ci

    @property
    def method_identifier(self) -> HostMethodIdentifier:
        """The method that needs to be invoked."""
        return self._ps_object.mi

    @property
    def method_parameters(self) -> typing.List[typing.Any]:
        """List of parameters to invoke the method with."""
        return _decode_host_call_parameters(self._ps_object.mi, self._ps_object.mp or [])

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} pipeline_id={self.pipeline_id!r} "
            f"ci={self.ci} method_identifier={self.method_identifier!r} method_parameters={self.method_parameters!r}>"
        )


@RegisterEvent
class PipelineHostResponseEvent(PSRPEvent[PipelineHostResponse, uuid.UUID]):
    """Pipeline Host Response Event.

    Event Information:

        Direction: Client -> Server

        For: Pipeline

        Message Type: :class:`psrpcore.types.PipelineHostResponse`

        Action Required: Send to higher layer for processing

    The client sends this message in response to a
    :class:`PipelineHostCallEvent`. It contains the host response information
    that is processed by the higher layer.
    """

    @property
    def ci(self) -> int:
        """The call id associated with the request."""
        return self._ps_object.ci

    @property
    def method_identifier(self) -> HostMethodIdentifier:
        """The method that was invoked."""
        return self._ps_object.mi

    @property
    def result(self) -> typing.Optional[typing.Any]:
        """The result (if any) from the host call."""
        return _decode_host_resp_result(self._ps_object.mi, getattr(self._ps_object, "mr", None))

    @property
    def error(self) -> typing.Optional[ErrorRecord]:
        """The error record if the host call failed."""
        me = getattr(self._ps_object, "me", None)
        if me and not isinstance(me, ErrorRecord):
            return ErrorRecord.FromPSObjectForRemoting(me)
        else:
            return me

    def __repr__(self) -> str:
        error = str(self.error) if self.error else None
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} pipeline_id={self.pipeline_id!r} "
            f"ci={self.ci} method_identifier={self.method_identifier!r} result={self.result!r} error={error!r}>"
        )


@RegisterEvent
class PipelineInputEvent(PSRPEvent[typing.Any, uuid.UUID]):
    """Pipeline Input Event.

    Event Information:

        Direction: Client -> Server

        For: Pipeline

        Message Type: Any

        Action Required: Send to higher layer for processing

    The client sends any input it desires for the pipeline. It is sent to the
    higher layer to be used with the running Pipeline.
    """

    @property
    def data(self) -> typing.Any:
        """The data sent as input."""
        return self._ps_object

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} pipeline_id={self.pipeline_id!r} "
            f"data={self.data!r}>"
        )


@RegisterEvent
class PipelineOutputEvent(PSRPEvent[typing.Any, uuid.UUID]):
    """Pipeline Output Event.

    Event Information:

        Direction: Server -> Client

        For: Pipeline

        Message Type: Any

        Action Required: Send to higher layer for processing

    The server sends this message when there is output generated from the
    Pipeline. The data should be shared with the higher layer as necessary.
    """

    @property
    def data(self) -> typing.Any:
        """The data sent as output."""
        return self._ps_object

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} pipeline_id={self.pipeline_id!r} "
            f"data={self.data!r}>"
        )


@RegisterEvent
class PipelineStateEvent(PSRPEvent[PipelineState, uuid.UUID]):
    """Pipeline State Event.

    Event Information:

        Direction: Server -> Client

        For: Pipeline

        Message Type: :class:`psrpcore.types.PipelineState`

        Action Required: Send to higher layer for processing

    The server sends this msg in response to a pipeline state change. The
    client should share this with the higher layer which checks whether it is
    ``Failed`` and use the reason to see why it did.
    """

    @property
    def state(self) -> PSInvocationState:
        """The Pipeline state."""
        return PSInvocationState(self._ps_object.PipelineState)

    @property
    def reason(self) -> typing.Optional[ErrorRecord]:
        """An error record containing the reason why the pipeline is ``Failed``."""
        return getattr(self._ps_object, "ExceptionAsErrorRecord", None)

    def __repr__(self) -> str:
        reason = str(self.reason) if self.reason else None
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} pipeline_id={self.pipeline_id!r} "
            f"state={self.state!r} reason={reason!r}>"
        )


@RegisterEvent
class ProgressRecordEvent(PSRPEvent[ProgressRecordMsg, _OptionalPipelineType]):
    """Progress Record Event.

    Event Information:

        Direction: Server -> Client

        For: Runspace Pool or Pipeline

        Message Type: :class:`psrpcore.types.ProgressRecordMsg`

        Action Required: Send to higher layer

    The server sends this message when there is a progress record generated by
    the server. The data should be shared with the higher layer as necessary.
    """

    @property
    def record(self) -> ProgressRecord:
        """The progress record emitted by the peer."""
        return ProgressRecord(
            Activity=self._ps_object.Activity,
            ActivityId=self._ps_object.ActivityId,
            CurrentOperation=self._ps_object.CurrentOperation,
            ParentActivityId=self._ps_object.ParentActivityId,
            PercentComplete=self._ps_object.PercentComplete,
            RecordType=self._ps_object.Type,
            SecondsRemaining=self._ps_object.SecondsRemaining,
            StatusDescription=self._ps_object.StatusDescription,
        )


@RegisterEvent
class PublicKeyEvent(PSRPEvent[PublicKey, None]):
    """Public Key Event.

    Event Information:

        Direction: Client -> Server

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.PublicKey`

        Action Required: Send response back to the client

    The client sends this message when it requests the encrypted session key
    used when serializing :class:`psrpcore.types.PSSecureString` objects. The
    server automatically generates the
    :class:`psrpcore.types.EncryptedSessionKey` as a response to the client.
    """

    @property
    def key(self) -> bytes:
        """The public key bytes from the peer."""
        return base64.b64decode(self._ps_object.PublicKey)


@RegisterEvent
class PublicKeyRequestEvent(PSRPEvent[PublicKeyRequest, None]):
    """Public Key Request Event.

    Event Information:

        Direction: Server -> Client

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.PublicKeyRequest`

        Action Required: Send response back to the server

    The server sends this message when it wants to serialize a
    :class:`psrpcore.types.PSSecureString` object. The client automatically
    generates the :class:`psrpcore.types.PublicKey` as a response to the
    server.
    """


@RegisterEvent
class ResetRunspaceStateEvent(PSRPEvent[ResetRunspaceState, None]):
    """Reset Runspace State Event.

    Event Information:

        Direction: Client -> Server

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.ResetRunspaceState`

        Action Required: Send to higher layer


    The client sends this message when it wants to reset the Runspace Pool
    state.
    """

    @property
    def ci(self) -> int:
        """The call id associated with the request."""
        return self._ps_object.ci

    def __repr__(self) -> str:
        return f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} ci={self.ci}>"


@RegisterEvent
class RunspaceAvailabilityEvent(PSRPEvent[RunspaceAvailability, None]):
    def __new__(
        cls,
        message_type: PSRPMessageType,
        ps_object: RunspaceAvailability,
        runspace_pool_id: uuid.UUID,
        pipeline_id: None = None,
    ) -> "RunspaceAvailabilityEvent":
        # Special case, this message has a boolean value when in response to Set[Max|Min]Runspaces and an Int64
        # value when in response to GetAvailableRunspaces. We want to make sure our event is clear what it is in
        # response to.
        if isinstance(ps_object.SetMinMaxRunspacesResponse, PSBool):
            return super().__new__(SetRunspaceAvailabilityEvent)
        else:
            return super().__new__(GetRunspaceAvailabilityEvent)

    @property
    def ci(self) -> int:
        """The call id associated with the request."""
        return self._ps_object.ci


class SetRunspaceAvailabilityEvent(RunspaceAvailabilityEvent):
    """Set Runspace Availability Event.

    Event Information:

        Direction: Server -> Client

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.RunspaceAvailability`

        Action Required: Send to higher layer

    The server sends this message in resposne to a
    :class:`psrpcore.types.SetMaxRunspaces`,
    :class:`psrpcore.types.SetMinRunspaces`, or
    :class:`psrpcore.types.ResetRunspaceState` message. The client
    automatically adjusts the runspace limits based on the response.
    """

    @property
    def success(self) -> bool:
        """Whether the request suceeded or not."""
        return self._ps_object.SetMinMaxRunspacesResponse

    def __repr__(self) -> str:
        return f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} ci={self.ci} success={self.success}>"


class GetRunspaceAvailabilityEvent(RunspaceAvailabilityEvent):
    """Get Runspace Availability Event.

    Event Information:

        Direction: Server -> Client

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.RunspaceAvailability`

        Action Required: Send to higher layer

    The server sends this message in resposne to a
    :class:`psrpcore.types.GetAvailableRunspaces` message.
    """

    @property
    def count(self) -> int:
        """The available runspaces specified by the server."""
        return int(self._ps_object.SetMinMaxRunspacesResponse)

    def __repr__(self) -> str:
        return f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} ci={self.ci} count={self.count}>"


@RegisterEvent
class RunspacePoolHostCallEvent(PSRPEvent[RunspacePoolHostCall, None]):
    """Runspace Pool Host Call Event.

    Event Information:

        Direction: Server -> Client

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.RunspacePoolHostCall`

        Action Required: Send to higher layer for processing

    The server sends this message when requesting the host to invoke a
    :class:`psrpcore.types.HostMethodIdentifier` method. It is up to the client
    to send this information to the higher layer and send back a response if
    required.
    """

    @property
    def ci(self) -> int:
        """The call id associated with the request."""
        return self._ps_object.ci

    @property
    def method_identifier(self) -> HostMethodIdentifier:
        """The method that needs to be invoked."""
        return self._ps_object.mi

    @property
    def method_parameters(self) -> typing.List[typing.Any]:
        """List of parameters to invoke the method with."""
        return _decode_host_call_parameters(self._ps_object.mi, self._ps_object.mp or [])

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} ci={self.ci} "
            f"method_identifier={self.method_identifier!r} method_parameters={self.method_parameters!r}>"
        )


@RegisterEvent
class RunspacePoolHostResponseEvent(PSRPEvent[RunspacePoolHostResponse, None]):
    """Runspace Pool Host Response Event.

    Event Information:

        Direction: Client -> Server

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.RunspacePoolHostResponse`

        Action Required: Send to higher layer for processing

    The client sends this message in response to a
    :class:`RunspacePoolHostCall`. It contains the host response information
    that is processed by the higher layer.
    """

    @property
    def ci(self) -> int:
        """The call id associated with the request."""
        return self._ps_object.ci

    @property
    def method_identifier(self) -> HostMethodIdentifier:
        """The method that was invoked."""
        return self._ps_object.mi

    @property
    def result(self) -> typing.Optional[typing.Any]:
        """The result (if any) from the host call."""
        return _decode_host_resp_result(self._ps_object.mi, getattr(self._ps_object, "mr", None))

    @property
    def error(self) -> typing.Optional[ErrorRecord]:
        """The error record if the host call failed."""
        return getattr(self._ps_object, "me", None)

    def __repr__(self) -> str:
        error = str(self.error) if self.error else None
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} ci={self.ci} "
            f"method_identifier={self.method_identifier!r} result={self.result!r} error={error!r}>"
        )


@RegisterEvent
class RunspacePoolInitDataEvent(PSRPEvent[RunspacePoolInitData, None]):
    """Runspace Pool Init Data Event

    Event Information:

        Direction: Server -> Client

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.`

        Action Required: Send to higher layer

    The server sends this message in response to the
    :class:`psrpcore.types.ConnectRunspacePool` message. The client adjusts the
    Runspace Pool min/max limits and the higher layer can process the data as
    required.
    """

    @property
    def max_runspaces(self) -> int:
        """The maximum number of runspaces in the Runspace Pool."""
        return self._ps_object.MaxRunspaces

    @property
    def min_runspaces(self) -> int:
        """The minimum number of runspaces in the Runspace Pool."""
        return self._ps_object.MinRunspaces

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} "
            f"min_runspaces={self.min_runspaces} max_runspaces={self.max_runspaces}>"
        )


@RegisterEvent
class RunspacePoolStateEvent(PSRPEvent[RunspacePoolStateMsg, None]):
    """Runspace Pool State Event.

    Event Information:

        Direction: Server -> Client

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.RunspacePoolStateMsg`

        Action Required: Send to higher layer for processing

    The server sends this msg in response to a runspace state change. The
    client should share this with the higher layer which checks whether it is
    ``Broken`` and use the reason to see why it did.
    """

    @property
    def state(self) -> RunspacePoolState:
        """The Runspace Pool state."""
        return RunspacePoolState(self._ps_object.RunspaceState)

    @property
    def reason(self) -> typing.Optional[ErrorRecord]:
        """An error record containing the reason why the runspace is ``Broken``."""
        return getattr(self._ps_object, "ExceptionAsErrorRecord", None)

    def __repr__(self) -> str:
        reason = str(self.reason) if self.reason else None
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} state={self.state!r} "
            f"reason={reason!r}>"
        )


@RegisterEvent
class SessionCapabilityEvent(PSRPEvent[SessionCapability, None]):
    """Session Capability Event.

    Event Information:

        Direction: Both

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.SessionCapability`

        Action Required: Nothing for client, send response back to the client

    The initial message received by both the client and server to state what
    protocol versions they understand. The server will automatically generate
    a :class:`psrpcore.types.SessionCapability` message that must be sent back
    to the client.
    """

    @property
    def ps_version(self) -> PSVersion:
        """The PowerShell version of the peer."""
        return self._ps_object.PSVersion

    @property
    def protocol_version(self) -> PSVersion:
        """The protocol version of the peer."""
        return self._ps_object.protocolversion

    @property
    def serialization_version(self) -> PSVersion:
        """The version of the peer's serializer."""
        return self._ps_object.SerializationVersion

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} ps_version={self.ps_version} "
            f"protocol_version={self.protocol_version} serialization_version={self.serialization_version}>"
        )


@RegisterEvent
class SetMaxRunspacesEvent(PSRPEvent[SetMaxRunspaces, None]):
    """Set Max Runspaces Event.

    Event Information:

        Direction: Client -> Server

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.SetMaxRunspaces`

        Action Required:

    The client sends this when trying to adjust the maximum available runspaces
    in a pool.
    """

    @property
    def ci(self) -> int:
        """The call id associated with the request."""
        return self._ps_object.ci

    @property
    def count(self) -> int:
        """The maximum runspace count to set."""
        return self._ps_object.MaxRunspaces

    def __repr__(self) -> str:
        return f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} ci={self.ci} count={self.count}>"


@RegisterEvent
class SetMinRunspacesEvent(PSRPEvent[SetMinRunspaces, None]):
    """Set Min Runspaces Event.

    Event Information:

        Direction: Client -> Server

        For: Runspace Pool

        Message Type: :class:`psrpcore.types.SetMinRunspaces`

        Action Required:

    The client sends this when trying to adjust the minimum available runspaces
    in a pool.
    """

    @property
    def ci(self) -> int:
        """The call id associated with the request."""
        return self._ps_object.ci

    @property
    def count(self) -> int:
        """The minimum runspace count to set."""
        return self._ps_object.MinRunspaces

    def __repr__(self) -> str:
        return f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} ci={self.ci} count={self.count}>"


@RegisterEvent
class UserEventEvent(PSRPEvent[UserEvent, None]):
    """User Event Event.

    Event Information:

        Direction: Server -> Client

        For: Runspace Pool or Pipeline

        Message Type: :class:`psrpcore.types.UserEvent`

        Action Required: Send to higher layer

    The server sends this message when reporting a user defined event from the
    runspace. The data should be shared with the higher layer as necessary.
    """

    @property
    def event(self) -> UserEvent:
        """The user event received from the peer."""
        return UserEvent.FromPSObjectForRemoting(self._ps_object)


@RegisterEvent
class VerboseRecordEvent(PSRPEvent[VerboseRecordMsg, _OptionalPipelineType]):
    """Verbose Record Event.

    Event Information:

        Direction: Server -> Client

        For: Runspace Pool or Pipeline

        Message Type: :class:`psrpcore.types.VerboseRecordMsg`

        Action Required: Send to higher layer

    The server sends this message when there is a verbose record generated by
    the server. The data should be shared with the higher layer as necessary.
    """

    @property
    def record(self) -> VerboseRecord:
        """The VerboseRecord emitted by the peer."""
        return self._ps_object

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} pipeline_id={self.pipeline_id!r} "
            f"record={self.record.Message!r}>"
        )


@RegisterEvent
class WarningRecordEvent(PSRPEvent[WarningRecordMsg, _OptionalPipelineType]):
    """Warning Record Event.

    Event Information:

        Direction: Server -> Client

        For: Runspace Pool or Pipeline

        Message Type: :class:`psrpcore.types.WarningRecordMsg`

        Action Required: Send to higher layer

    The server sends this message when there is a warning record generated by
    the server. The data should be shared with the higher layer as necessary.
    """

    @property
    def record(self) -> WarningRecord:
        """The WarningRecord emitted by the peer."""
        return self._ps_object

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} runspace_pool_id={self.runspace_pool_id!r} pipeline_id={self.pipeline_id!r} "
            f"record={self.record.Message!r}>"
        )



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

"""PSRP/.NET Host Types.

The PSRP/.NET host type and metadata definitions. Contains information used
for host methods used in PSRP.
"""

import typing

from psrpcore.types._base import PSNoteProperty, PSObject, PSType, add_note_property
from psrpcore.types._collection import PSList
from psrpcore.types._complex import ConsoleColor
from psrpcore.types._enum import PSEnumBase, PSFlagBase
from psrpcore.types._primitive import PSBool, PSChar, PSInt, PSString


@PSType(["System.Management.Automation.Remoting.RemoteHostMethodId"])
class HostMethodIdentifier(PSEnumBase):
    """Host Method Identifier enum.

    This is an enum class for the
    System.Management.Automation.Remoting.RemoteHostMethodId .NET class. It is
    documented in PSRP under `[MS-PSRP] 2.2.3.17 Host Method Identifier`_.

    The values are used in :class:`psrpcore.types.RunspacePoolHostCall` and
    :class:`psrpcore.types.PipelineHostCall` to identify what method should be
    invoked.

    .. _[MS-PSRP] 2.2.3.17 Host Method Identifier:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/ddd2a4d1-797d-4d73-8372-7a77a62fb204
    """

    GetName = 1
    GetVersion = 2
    GetInstanceId = 3
    GetCurrentCulture = 4
    GetCurrentUICulture = 5
    SetShouldExit = 6
    EnterNestedPrompt = 7
    ExitNestedPrompt = 8
    NotifyBeginApplication = 9
    NotifyEndApplication = 10
    ReadLine = 11
    ReadLineAsSecureString = 12
    Write1 = 13
    Write2 = 14
    WriteLine1 = 15
    WriteLine2 = 16
    WriteLine3 = 17
    WriteErrorLine = 18
    WriteDebugLine = 19
    WriteProgress = 20
    WriteVerboseLine = 21
    WriteWarningLine = 22
    Prompt = 23
    PromptForCredential1 = 24
    PromptForCredential2 = 25
    PromptForChoice = 26
    GetForegroundColor = 27
    SetForegroundColor = 28
    GetBackgroundColor = 29
    SetBackgroundColor = 30
    GetCursorPosition = 31
    SetCursorPosition = 32
    GetWindowPosition = 33
    SetWindowPosition = 34
    GetCursorSize = 35
    SetCursorSize = 36
    GetBufferSize = 37
    SetBufferSize = 38
    GetWindowSize = 39
    SetWindowSize = 40
    GetWindowTitle = 41
    SetWindowTitle = 42
    GetMaxWindowSize = 43
    GetMaxPhysicalWindowSize = 44
    GetKeyAvailable = 45
    ReadKey = 46
    FlushInputBuffer = 47
    SetBufferContents1 = 48
    SetBufferContents2 = 49
    GetBufferContents = 50
    ScrollBufferContents = 51
    PushRunspace = 52
    PopRunspace = 53
    GetIsRunspacePushed = 54
    GetRunspace = 55
    PromptForChoiceMultipleSelection = 56

    def is_void(self) -> bool:
        """Whether the method is a void method and doesn't return data."""
        return self in [
            HostMethodIdentifier.SetShouldExit,
            HostMethodIdentifier.EnterNestedPrompt,
            HostMethodIdentifier.ExitNestedPrompt,
            HostMethodIdentifier.NotifyBeginApplication,
            HostMethodIdentifier.NotifyEndApplication,
            HostMethodIdentifier.PushRunspace,
            HostMethodIdentifier.PopRunspace,
            HostMethodIdentifier.Write1,
            HostMethodIdentifier.Write2,
            HostMethodIdentifier.WriteLine1,
            HostMethodIdentifier.WriteLine2,
            HostMethodIdentifier.WriteLine3,
            HostMethodIdentifier.WriteErrorLine,
            HostMethodIdentifier.WriteDebugLine,
            HostMethodIdentifier.WriteProgress,
            HostMethodIdentifier.WriteVerboseLine,
            HostMethodIdentifier.WriteWarningLine,
            HostMethodIdentifier.SetForegroundColor,
            HostMethodIdentifier.SetBackgroundColor,
            HostMethodIdentifier.SetCursorPosition,
            HostMethodIdentifier.SetWindowPosition,
            HostMethodIdentifier.SetCursorSize,
            HostMethodIdentifier.SetBufferSize,
            HostMethodIdentifier.SetWindowSize,
            HostMethodIdentifier.SetWindowTitle,
            HostMethodIdentifier.FlushInputBuffer,
            HostMethodIdentifier.SetBufferContents1,
            HostMethodIdentifier.SetBufferContents2,
            HostMethodIdentifier.ScrollBufferContents,
        ]


@PSType(["System.Management.Automation.Host.ControlKeyStates"])
class ControlKeyStates(PSFlagBase):
    """System.Management.Automation.Host.ControlKeyStates enum flags.

    Defines the state of the control key. It is documented in PSRP under
    `[MS-PSRP] 2.2.3.27 ControlKeyStates`_.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.Host.ControlKeyStates`_ .NET class.

    .. _[MS-PSRP] 2.2.3.27 ControlKeyStates:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/bd7241a2-4ba0-4db1-a2b3-77ea1a8a4cbf

    .. _System.Management.Automation.Host.ControlKeyStates:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.controlkeystates
    """

    none = 0  #: No control key state.
    RightAltPressed = 1  #: The right alt key is pressed.
    LeftAltPressed = 2  #: The left alt key is pressed.
    RightCtrlPressed = 4  #: The right ctrl key is pressed.
    LeftCtrlPressed = 8  #: The left ctrl key is pressed.
    ShiftPressed = 16  #: The shift key is pressed.
    NumLockOn = 32  #: The numlock light is on.
    ScrollLockOn = 64  #: The scrolllock light is on.
    CapsLockOn = 128  #: The capslock light is on.
    EnhancedKey = 256  #: The key is enhanced.


@PSType(["System.Management.Automation.Host.ControlKeyStates"])
class BufferCellType(PSFlagBase):
    """System.Management.Automation.Host.BufferCellType enum flags.

    Defines three types of BufferCells to accommodate for hosts that use up to
    two cells to display a character in some languages such as Chinese and
    Japanese. It is documented in PSRP under
    `[MS-PSRP] 2.2.3.29 BufferCellType`_.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.Host.BufferCellType`_ .NET class.

    .. _[MS-PSRP] 2.2.3.29 BufferCellType:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/99938ede-6d84-422e-b75d-ace93ea85ea2

    .. _System.Management.Automation.Host.BufferCellType:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.buffercelltype
    """

    Complete = 0  #: Character occupies one BufferCell.
    Leading = 1  #: Character occupies two BufferCells and this is the leading one.
    Trailing = 2  #: Preceded by a Leading BufferCell.


@PSType(["System.Management.Automation.Host.ReadKeyOptions"])
class ReadKeyOptions(PSFlagBase):
    """System.Management.Automation.Host.ReadKeyOptions enum flags.

    Governs the behavior of :class:`HostMethodIdentifier.ReadKey`.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.Host.ReadKeyOptions`_ .NET class.

    .. _System.Management.Automation.Host.ReadKeyOptions:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.readkeyoptions
    """

    AllowCtrlC = 1  #: Allow Ctrl-C to be processed as a keystroke, as opposed to causing a break event.
    NoEcho = 2  #: Do not display the character for the key in the window when pressed.
    IncludeKeyDown = 4  #: Include key down events. One of IncludeKeyDown and IncludeKeyUp or both must be specified.
    IncludeKeyUp = 8  #: Include key up events. One of IncludeKeyDown and IncludeKeyUp or both must be specified.


@PSType(
    type_names=[
        "System.Management.Automation.Host.BufferCell",
        "System.ValueType",
    ],
    adapted_properties=[
        PSNoteProperty("Character", ps_type=PSChar),
        PSNoteProperty("ForegroundColor", ps_type=ConsoleColor),
        PSNoteProperty("BackgroundColor", ps_type=ConsoleColor),
        PSNoteProperty("BufferCellType", ps_type=BufferCellType),
    ],
)
class BufferCell(PSObject):
    """BufferCell.

    Represents the contents of a cell of a host's screen buffer. It is
    documented in PSRP under `[MS-PSRP] 2.2.3.28 BufferCell`_. but the PSRP
    documentation represents how this value is serialized under
    host method invocations not the .NET type.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.Host.BufferCell`_ .NET class.

    Args:
        Character: Character visible in the cell.
        ForegroundColor: Foreground color.
        BackgroundColor: Background color.
        BufferCellType: Type of the buffer cell.

    .. _[MS-PSRP] 2.2.3.28 BufferCell:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/d6270c27-8855-46b6-834c-5a5d188bfe70

    .. _System.Management.Automation.Host.BufferCell:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.buffercell
    """

    @classmethod
    def FromPSObjectForRemoting(
        cls,
        obj: PSObject,
        **kwargs: typing.Any,
    ) -> "BufferCell":
        if isinstance(obj, BufferCell):
            return obj

        return BufferCell(
            Character=obj.character,
            ForegroundColor=ConsoleColor(obj.foregroundColor),
            BackgroundColor=ConsoleColor(obj.backgroundColor),
            BufferCellType=BufferCellType(obj.bufferCellType),
        )

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "BufferCell",
        for_host: bool = False,
        **kwargs: typing.Any,
    ) -> PSObject:
        if not for_host:
            return instance

        obj = PSObject()
        obj.PSObject.type_names = []
        add_note_property(obj, "character", instance.Character, ps_type=PSChar)
        add_note_property(obj, "foregroundColor", instance.ForegroundColor.value, ps_type=PSInt)
        add_note_property(obj, "backgroundColor", instance.BackgroundColor.value, ps_type=PSInt)
        add_note_property(obj, "bufferCellType", instance.BufferCellType.value, ps_type=PSInt)

        return obj


@PSType(
    type_names=["System.Management.Automation.Host.ChoiceDescription"],
    adapted_properties=[
        PSNoteProperty("Label", mandatory=True, ps_type=PSString),
        PSNoteProperty("HelpMessage", ps_type=PSString),
    ],
)
class ChoiceDescription(PSObject):
    """ChoiceDescription.

    Provides a description of a choice for use by
    :class:`HostMethodIdentifier.PromptForChoice`.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.Host.ChoiceDescription`_ .NET class.

    Args:
        Label: Human friendly message to describe the choice.
        HelpMessage: Help details of the choice.

    .. _System.Management.Automation.Host.ChoiceDescription:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.choicedescription
    """

    @classmethod
    def FromPSObjectForRemoting(
        cls,
        obj: PSObject,
        **kwargs: typing.Any,
    ) -> "ChoiceDescription":
        if isinstance(obj, ChoiceDescription):
            return obj

        return ChoiceDescription(Label=obj.label, HelpMessage=obj.helpMessage)

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "ChoiceDescription",
        for_host: bool = False,
        **kwargs: typing.Any,
    ) -> PSObject:
        if not for_host:
            return instance

        obj = PSObject()
        obj.PSObject.type_names = []
        add_note_property(obj, "label", instance.Label, ps_type=PSString)
        add_note_property(obj, "helpMessage", instance.HelpMessage, ps_type=PSString)

        return obj


@PSType(
    type_names=[
        "System.Management.Automation.Host.Coordinates",
        "System.ValueType",
    ],
    adapted_properties=[
        PSNoteProperty("X", mandatory=True, ps_type=PSInt),
        PSNoteProperty("Y", mandatory=True, ps_type=PSInt),
    ],
)
class Coordinates(PSObject):
    """Coordinates

    Represents an x,y coordinate pair. It is documented under
    `[MS-PSRP] 2.2.3.1 Coordinates`_ but the PSRP documentation represents how
    this value is serialized under :class:`HostDefaultData` not the .NET type.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.Host.Coordinates`_ .NET class.

    Args:
        X: X coordinate (0 is the leftmost column).
        Y: Y coordinate (0 is the topmost row).

    .. _[MS-PSRP] 2.2.3.1 Coordinates:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/05db8994-ec5c-485c-9e91-3a398e461d38

    .. _System.Management.Automation.Host.Coordinates:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.coordinates
    """

    @classmethod
    def FromPSObjectForRemoting(
        cls,
        obj: PSObject,
        **kwargs: typing.Any,
    ) -> "Coordinates":
        if isinstance(obj, Coordinates):
            return obj

        return Coordinates(X=obj.x, Y=obj.y)

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "Coordinates",
        for_host: bool = False,
        **kwargs: typing.Any,
    ) -> PSObject:
        if not for_host:
            return instance

        obj = PSObject()
        obj.PSObject.type_names = []
        add_note_property(obj, "x", instance.X, ps_type=PSInt)
        add_note_property(obj, "y", instance.Y, ps_type=PSInt)

        return obj


@PSType(
    type_names=["System.Management.Automation.Host.FieldDescription"],
    adapted_properties=[
        PSNoteProperty("Name", mandatory=True, ps_type=PSString),
        PSNoteProperty("ParameterTypeName", ps_type=PSString),
        PSNoteProperty("ParameterTypeFullName", ps_type=PSString),
        PSNoteProperty("ParameterAssemblyFullName", ps_type=PSString),
        PSNoteProperty("Label", value="", ps_type=PSString),
        PSNoteProperty("HelpMessage", value="", ps_type=PSString),
        PSNoteProperty("IsMandatory", value=True, ps_type=PSBool),
        PSNoteProperty("DefaultValue", value=None),
        PSNoteProperty("Attributes", ps_type=PSList),
    ],
)
class FieldDescription(PSObject):
    """FieldDescription.

    Provides a description of a field for use by
    :class:`HostMethodIdentifier.Prompt`.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.Host.FieldDescription`_ .NET class.

    Args:
        Name: The name of the field.
        ParameterTypeName: Short name of the field value type.
        ParameterTypeFullName: Full name of the field value type.
        ParameterAssemblyFullName: .NET assembly name of the field value type.
        Label: Human friendly message to describe the field.
        HelpMessage: Help details of the field.
        IsMandatory: The value must be supplied for this field.
        DefaultValue: The default value, if any, for the field.
        Attributes: List of attributes that apply to the field.

    .. _System.Management.Automation.Host.FieldDescription:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.fielddescription
    """

    @classmethod
    def FromPSObjectForRemoting(
        cls,
        obj: PSObject,
        **kwargs: typing.Any,
    ) -> "FieldDescription":
        if isinstance(obj, FieldDescription):
            return obj

        return FieldDescription(
            Name=obj.name,
            ParameterTypeName=obj.parameterTypeName,
            ParameterTypeFullName=obj.parameterTypeFullName,
            ParameterAssemblyFullName=obj.parameterAssemblyFullName,
            Label=obj.label,
            HelpMessage=obj.helpMessage,
            IsMandatory=obj.isMandatory,
            Attributes=obj.metadata,
        )

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "FieldDescription",
        for_host: bool = False,
        **kwargs: typing.Any,
    ) -> PSObject:
        if not for_host:
            return instance

        obj = PSObject()
        obj.PSObject.type_names = []
        add_note_property(obj, "name", instance.Name, ps_type=PSString)
        add_note_property(obj, "label", instance.Label, ps_type=PSString)
        add_note_property(obj, "parameterTypeName", instance.ParameterTypeName, ps_type=PSString)
        add_note_property(obj, "parameterTypeFullName", instance.ParameterTypeFullName, ps_type=PSString)
        add_note_property(obj, "parameterAssemblyFullName", instance.ParameterAssemblyFullName, ps_type=PSString)
        add_note_property(obj, "helpMessage", instance.HelpMessage, ps_type=PSString)
        add_note_property(obj, "isMandatory", instance.IsMandatory, ps_type=PSBool)
        add_note_property(obj, "metadata", [])
        add_note_property(obj, "modifiedByRemotingProtocol", False, ps_type=PSBool)
        add_note_property(obj, "isFromRemoteHost", False, ps_type=PSBool)

        return obj


@PSType(
    type_names=["System.Management.Automation.Host.KeyInfo"],
    adapted_properties=[
        PSNoteProperty("VirtualKeyCode", ps_type=PSInt),
        PSNoteProperty("Character", ps_type=PSChar),
        PSNoteProperty("ControlKeyState", ps_type=ControlKeyStates),
        PSNoteProperty("KeyDown", ps_type=PSBool),
    ],
)
class KeyInfo(PSObject):
    """KeyInfo.

    Represents information of a keystroke. It is documented in PSRP under
    `[MS-PSRP] 2.2.3.26 KeyInfo`_.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.Host.KeyInfo`_ .NET class.

    Args:
        VirtualKeyCode: A virtual key code that identifies the given key in a
            device-independent manner.
        Character: Character corresponding to the pressed keys.
        ControlKeyState: State of the control keys.
        KeyDown: True if the event was generated when a key was pressed.

    .. _[MS-PSRP] 2.2.3.26 KeyInfo:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/481442e2-5304-4679-b16d-6e53c351339d

    .. _System.Management.Automation.Host.KeyInfo:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.keyinfo
    """

    @classmethod
    def FromPSObjectForRemoting(
        cls,
        obj: PSObject,
        **kwargs: typing.Any,
    ) -> "KeyInfo":
        if isinstance(obj, KeyInfo):
            return obj

        return KeyInfo(
            VirtualKeyCode=obj.virtualKeyCode,
            Character=obj.character,
            ControlKeyState=ControlKeyStates(obj.controlKeyState),
            KeyDown=obj.keyDown,
        )

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "KeyInfo",
        for_host: bool = False,
        **kwargs: typing.Any,
    ) -> PSObject:
        if not for_host:
            return instance

        obj = PSObject()
        obj.PSObject.type_names = []
        add_note_property(obj, "virtualKeyCode", instance.VirtualKeyCode, ps_type=PSInt)
        add_note_property(obj, "character", instance.Character, ps_type=PSChar)
        add_note_property(obj, "controlKeyState", instance.ControlKeyState.value, ps_type=PSInt)
        add_note_property(obj, "keyDown", instance.KeyDown, ps_type=PSBool)

        return obj


@PSType(
    type_names=[
        "System.Management.Automation.Host.Rectangle",
        "System.ValueType",
    ],
    adapted_properties=[
        PSNoteProperty("Left", mandatory=True, ps_type=PSInt),
        PSNoteProperty("Top", mandatory=True, ps_type=PSInt),
        PSNoteProperty("Right", mandatory=True, ps_type=PSInt),
        PSNoteProperty("Bottom", mandatory=True, ps_type=PSInt),
    ],
)
class Rectangle(PSObject):
    """Rectangle.

    Represents a rectangular region of the screen.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.Host.Rectangle`_ .NET class.

    Args:
        Left: Gets and sets the left side of the rectangle.
        Top: Gets and sets the top of the rectangle.
        Right: Gets and sets the right side of the rectangle.
        Bottom: Gets and sets the bottom of the rectanngle.

    .. _System.Management.Automation.Host.Rectangle:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.rectangle
    """

    @classmethod
    def FromPSObjectForRemoting(
        cls,
        obj: PSObject,
        **kwargs: typing.Any,
    ) -> "Rectangle":
        if isinstance(obj, Rectangle):
            return obj

        return Rectangle(Left=obj.left, Top=obj.top, Right=obj.right, Bottom=obj.bottom)

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "Rectangle",
        for_host: bool = False,
        **kwargs: typing.Any,
    ) -> PSObject:
        if not for_host:
            return instance

        obj = PSObject()
        obj.PSObject.type_names = []
        add_note_property(obj, "left", instance.Left, ps_type=PSInt)
        add_note_property(obj, "top", instance.Top, ps_type=PSInt)
        add_note_property(obj, "right", instance.Right, ps_type=PSInt)
        add_note_property(obj, "bottom", instance.Bottom, ps_type=PSInt)

        return obj


@PSType(
    type_names=[
        "System.Management.Automation.Host.Size",
        "System.ValueType",
    ],
    adapted_properties=[
        PSNoteProperty("Width", mandatory=True, ps_type=PSInt),
        PSNoteProperty("Height", mandatory=True, ps_type=PSInt),
    ],
)
class Size(PSObject):
    """Size

    Represents a width and height pair. It is documented under
    `[MS-PSRP] 2.2.3.2 Size`_ but the PSRP documentation represents how this
    value is serialized under :class:`HostDefaultData` not the .NET type.


    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.Host.Size`_ .NET class.

    Args:
        Width: The width of an area.
        Height: The height of an area.

    .. _[MS-PSRP] 2.2.3.2 Size:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/98cd950f-cc12-4ab4-955d-c389e3089856

    .. _System.Management.Automation.Host.Size:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.size
    """

    @classmethod
    def FromPSObjectForRemoting(
        cls,
        obj: PSObject,
        **kwargs: typing.Any,
    ) -> "Size":
        if isinstance(obj, Size):
            return obj

        return Size(Width=obj.width, Height=obj.height)

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "Size",
        for_host: bool = False,
        **kwargs: typing.Any,
    ) -> PSObject:
        if not for_host:
            return instance

        obj = PSObject()
        obj.PSObject.type_names = []
        add_note_property(obj, "width", instance.Width, ps_type=PSInt)
        add_note_property(obj, "height", instance.Height, ps_type=PSInt)

        return obj


@PSType(
    extended_properties=[
        PSNoteProperty("ForegroundColor", mandatory=True, ps_type=ConsoleColor),
        PSNoteProperty("BackgroundColor", mandatory=True, ps_type=ConsoleColor),
        PSNoteProperty("CursorPosition", mandatory=True, ps_type=Coordinates),
        PSNoteProperty("WindowPosition", mandatory=True, ps_type=Coordinates),
        PSNoteProperty("CursorSize", mandatory=True, ps_type=PSInt),
        PSNoteProperty("BufferSize", mandatory=True, ps_type=Size),
        PSNoteProperty("WindowSize", mandatory=True, ps_type=Size),
        PSNoteProperty("MaxWindowSize", mandatory=True, ps_type=Size),
        PSNoteProperty("MaxPhysicalWindowSize", mandatory=True, ps_type=Size),
        PSNoteProperty("WindowTitle", mandatory=True, ps_type=PSString),
    ],
    skip_inheritance=True,
)
class HostDefaultData(PSObject):
    """HostInfo default data.

    This defines the default data for a PSHost when creating a RunspacePool or
    Pipeline. This does not represent an actual .NET type but is an internal
    object representation used by PSRP itself. This type represents the
    `hostDefaultData` property documented at `[MS-PSRP] 2.2.3.14 HostInfo`_.

    Args:
        ForegroundColor: Color of the character on the screen buffer.
        BackgroundColor: Color behind characters on the screen buffer.
        CursorPosition: Cursor position in the screen buffer.
        WindowPosition: Position of the view window relative to the screen
            buffer.
        CursorSize: Cursor size as a percentage 0..100.
        BufferSize: Current size of the screen buffer, measured in character
            cells.
        WindowSize: Current view window size, measured in character cells.
        MaxWindowSize:  Size of the largest window position for the current
            buffer.
        MaxPhysicalWindowSize: Largest window possible ignoring the current
            buffer dimensions.
        WindowTitle: The titlebar text of the current view window.

    .. _[MS-PSRP] 2.2.3.14 HostInfo:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/510fd8f3-e3ac-45b4-b622-0ad5508a5ac6
    """

    @classmethod
    def FromPSObjectForRemoting(
        cls,
        obj: PSObject,
        **kwargs: typing.Any,
    ) -> "HostDefaultData":
        return HostDefaultData(
            ForegroundColor=ConsoleColor(obj.data[0].V),
            BackgroundColor=ConsoleColor(obj.data[1].V),
            CursorPosition=Coordinates.FromPSObjectForRemoting(obj.data[2].V),
            WindowPosition=Coordinates.FromPSObjectForRemoting(obj.data[3].V),
            CursorSize=obj.data[4].V,
            BufferSize=Size.FromPSObjectForRemoting(obj.data[5].V),
            WindowSize=Size.FromPSObjectForRemoting(obj.data[6].V),
            MaxWindowSize=Size.FromPSObjectForRemoting(obj.data[7].V),
            MaxPhysicalWindowSize=Size.FromPSObjectForRemoting(obj.data[8].V),
            WindowTitle=obj.data[9].V,
        )

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "HostDefaultData",
        **kwargs: typing.Any,
    ) -> PSObject:
        obj = PSObject()

        def dict_value(value: typing.Union[int, str, PSObject], value_type: str) -> PSObject:
            dict_obj = PSObject()
            add_note_property(dict_obj, "T", value_type, ps_type=PSString)
            add_note_property(dict_obj, "V", value)
            return dict_obj

        data = {}
        for idx, prop in enumerate(instance.PSObject.extended_properties):
            value = prop.get_value(instance)
            if isinstance(value, ConsoleColor):
                value = dict_value(PSInt(value), value.PSObject.type_names[0])

            elif isinstance(value, Coordinates):
                raw = Coordinates.ToPSObjectForRemoting(value, for_host=True)
                value = dict_value(raw, value.PSObject.type_names[0])

            elif isinstance(value, Size):
                raw = Size.ToPSObjectForRemoting(value, for_host=True)
                value = dict_value(raw, value.PSObject.type_names[0])

            else:
                value = dict_value(value, value.PSObject.type_names[0])

            data[idx] = value

        add_note_property(obj, "data", data)

        return obj


@PSType(
    extended_properties=[
        PSNoteProperty("IsHostNull", value=True, ps_type=PSBool),
        PSNoteProperty("IsHostUINull", value=True, ps_type=PSBool),
        PSNoteProperty("IsHostRawUINull", value=True, ps_type=PSBool),
        PSNoteProperty("UseRunspaceHost", value=True, ps_type=PSBool),
        PSNoteProperty("HostDefaultData", ps_type=HostDefaultData),
    ],
    skip_inheritance=True,
)
class HostInfo(PSObject):
    """HostInfo.

    Defines the PSHost information. Message is defined in
    `[MS-PSRP] 2.2.3.14 HostInfo`_.

    Args:
        IsHostNull: Whether there is a PSHost ``False`` or not ``True``.
        IsHostUINull: Whether the PSHost implements the `UI` implementation
            methods ``False`` or not ``True``.
        IsHostRawUINull: Whether the PSHost UI implements the ``RawUI``
            implementation methods ``False`` or not ``True``.
        UseRunspaceHost: When creating a pipeline, set this to ``True`` to get
            it to use the associated RunspacePool host.
        HostDefaultData: Host default data associated with the PSHost
            implementation. Can be ``None`` if not implemented.

    .. _[MS-PSRP] 2.2.3.14 HostInfo:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/510fd8f3-e3ac-45b4-b622-0ad5508a5ac6
    """

    @classmethod
    def FromPSObjectForRemoting(
        cls,
        obj: PSObject,
        **kwargs: typing.Any,
    ) -> "HostInfo":
        """Convert the raw HostInfo PSObject back to this easier to use object."""
        host_data = getattr(obj, "_hostDefaultData", None)
        if host_data is not None:
            host_data = HostDefaultData.FromPSObjectForRemoting(host_data)

        return HostInfo(
            IsHostNull=obj._isHostNull,
            IsHostUINull=obj._isHostUINull,
            IsHostRawUINull=obj._isHostRawUINull,
            UseRunspaceHost=obj._useRunspaceHost,
            HostDefaultData=host_data,
        )

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "HostInfo",
        **kwargs: typing.Any,
    ) -> PSObject:
        obj = PSObject()
        add_note_property(obj, "_isHostNull", instance.IsHostNull)
        add_note_property(obj, "_isHostUINull", instance.IsHostUINull)
        add_note_property(obj, "_isHostRawUINull", instance.IsHostRawUINull)
        add_note_property(obj, "_useRunspaceHost", instance.UseRunspaceHost)

        if instance.HostDefaultData:
            add_note_property(obj, "_hostDefaultData", instance.HostDefaultData)

        return obj
